# Purpose: Module for writing DAS data.
# Author: Minzhe Hu
# Date: 2024.11.20
# Email: hmz2018@mail.ustc.edu.cn
import os
import warnings
import pickle
import numpy as np
import h5py
import segyio
from shutil import copyfile
from nptdms import TdmsFile, TdmsWriter, RootObject, GroupObject, ChannelObject
from datetime import datetime


def write(sec, fname, ftype=None, raw_fname=None):
    fun_map = {'tdms': _write_tdms, 'h5': _write_h5, 'sgy': _write_segy}
    if ftype is None:
        ftype = str(fname).lower().split('.')[-1]
    ftype.replace('hdf5', 'h5')
    ftype.replace('segy', 'sgy')
    if ftype == 'pkl':
        write_pkl(sec, fname)
    elif ftype == 'npy':
        np.save(fname, sec.data)
    else:
        fun_map[ftype](sec, fname, raw_fname=raw_fname)
    return None


def write_pkl(sec, fname):
    with open(fname, 'wb') as f:
        pickle.dump(sec.__dict__, f)
    return None


def _write_tdms(sec, fname, raw_fname=None):
    if raw_fname is None:
        key = 'Measurement'
        file_prop = {}
        group_prop = {}
    else:
        original_file = TdmsFile(raw_fname)
        group_name = [group.name for group in original_file.groups()]
        if 'Measurement' in group_name:
            key = 'Measurement'
        elif 'DAS' in group_name:
            key = 'DAS'
        else:
            key = group_name[0]
        file_prop = original_file.properties
        group_prop = original_file[key].properties

    if 'Spatial Resolution' in group_prop.keys():
        group_prop['Spatial Resolution'] = sec.dx
    else:
        file_prop['SpatialResolution[m]'] = sec.dx

    if 'Time Base' in group_prop.keys():
        group_prop['Time Base'] = 1. / sec.fs
    else:
        file_prop['SamplingFrequency[Hz]'] = sec.fs

    if 'Total Channels' in group_prop.keys():
        group_prop['Total Channels'] = sec.nch

    if 'Initial Channel' in group_prop.keys():
        group_prop['Initial Channel'] = sec.start_channel

    file_prop['Start Distance (m)'] = sec.start_distance
    if isinstance(sec.start_time, datetime):
        start_time = sec.start_time
    else:
        start_time = datetime.fromtimestamp(sec.start_time)

    if raw_fname is None:
        file_prop['ISO8601 Timestamp'] = start_time.strftime(
            '%Y-%m-%dT%H:%M:%S.%f%z')
        group_prop['Trigger Time'] = np.datetime64(start_time.remove_tz())
    else:
        if 'ISO8601 Timestamp' in file_prop.keys():
            file_prop['ISO8601 Timestamp'] = start_time.strftime(
                '%Y-%m-%dT%H:%M:%S.%f%z')
        else:
            for s in ['GPSTimeStamp', 'CPUTimeStamp', 'Trigger Time']:
                if s in group_prop.keys():
                    group_prop[s] = np.datetime64(start_time.remove_tz())
                    break

    if hasattr(sec, 'gauge_length'):
        file_prop['GaugeLength'] = sec.gauge_length

    with TdmsWriter(fname) as tdms_file:
        root_object = RootObject(file_prop)
        group_object = GroupObject(key, properties=group_prop)
        if raw_fname and len(original_file[key]) == 1:
            channel = ChannelObject(key, original_file[key].channels()[0].name,
                                    sec.data.T.flatten(), properties={})
            tdms_file.write_segment([root_object, group_object, channel])
        else:
            channel_list = []
            for ch, d in enumerate(sec.data):
                channel_list.append(ChannelObject(key,
                                                  str(ch + sec.start_channel),
                                                  d, properties={}))

            tdms_file.write_segment([root_object, group_object] + channel_list)
    return None


def _update_h5_dataset(h5_file, path, name, data):
    attrs = h5_file[path + name].attrs
    del h5_file[path + name]
    h5_file.get(path).create_dataset(name, data=data)
    for key, value in attrs.items():
        h5_file[path + name].attrs[key] = value
    return None


def _write_h5(sec, fname, raw_fname=None):
    if raw_fname is None:
        with h5py.File(fname, 'w') as h5_file:
            h5_file.create_group('Acquisition/Raw[0]')
            h5_file.get('Acquisition/Raw[0]/').\
                create_dataset('RawData', data=sec.data)
            if isinstance(sec.start_time, datetime):
                h5_file['Acquisition/Raw[0]/RawData'].attrs['PartStartTime'] = \
                    np.bytes_(
                    sec.start_time.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))
                stime = sec.start_time.timestamp() * 1e6
                DataTime = np.arange(
                    stime, stime + sec.nt / sec.fs, 1 / sec.fs)
            else:
                h5_file['Acquisition/Raw[0]/RawData'].attrs['PartStartTime'] = \
                    np.bytes_(str(sec.start_time))
                DataTime = sec.start_time + np.arange(0, sec.nt / sec.fs,
                                                      1 / sec.fs)

            h5_file.get('Acquisition/Raw[0]/').\
                create_dataset('RawDataTime', data=DataTime)
            h5_file['Acquisition/Raw[0]'].attrs['OutputDataRate'] = sec.fs
            h5_file['Acquisition'].attrs['SpatialSamplingInterval'] = sec.dx
            if hasattr(sec, 'gauge_length'):
                h5_file['Acquisition'].attrs['GaugeLength'] = sec.gauge_length
            else:
                h5_file['Acquisition'].attrs['GaugeLength'] = np.nan
    else:
        if not os.path.exists(fname) or not os.path.samefile(raw_fname, fname):
            copyfile(raw_fname, fname)
        with h5py.File(fname, 'r+') as h5_file:
            group = list(h5_file.keys())[0]
            if len(h5_file.keys()) == 10:
                if h5_file['header/dimensionNames'][0] == b'time':
                    _update_h5_dataset(h5_file, '/', 'data', sec.data.T)
                elif h5_file['header/dimensionNames'][0] == b'distance':
                    _update_h5_dataset(h5_file, '/', 'data', sec.data)

                _update_h5_dataset(h5_file, 'header', 'dx', sec.dx)
                _update_h5_dataset(h5_file, 'header', 'dt', 1 / sec.fs)
                if isinstance(sec.start_time, datetime):
                    _update_h5_dataset(h5_file, 'header', 'time',
                                       sec.start_time.timestamp())
                else:
                    _update_h5_dataset(h5_file, 'header', 'time',
                                       sec.start_time)
                if hasattr(sec, 'gauge_length'):
                    _update_h5_dataset(h5_file, '/', 'gaugeLength',
                                       sec.gauge_length)
                if hasattr(sec, 'scale'):
                    _update_h5_dataset(h5_file, '/', 'dataScale', sec.scale)
            elif len(h5_file.keys()) == 5:
                _update_h5_dataset(h5_file, '/', 'strain', sec.data.T)
                _update_h5_dataset(h5_file, '/', 'spatialsampling', sec.dx)
                _update_h5_dataset(h5_file, '/', 'RepetitionFrequency', sec.fs)
                if hasattr(sec, 'gauge_length'):
                    _update_h5_dataset(h5_file, '/', 'GaugeLength',
                                       sec.gauge_length)
            elif group == 'Acquisition':
                h5_file['Acquisition'].attrs['NumberOfLoci'] = sec.nch
                _update_h5_dataset(h5_file, 'Acquisition/Raw[0]/', 'RawData',
                                   sec.data)
                if isinstance(sec.start_time, datetime):
                    if isinstance(h5_file['Acquisition/Raw[0]/RawData'].
                                  attrs['PartStartTime'], bytes):
                        h5_file['Acquisition/Raw[0]/RawData'].\
                            attrs['PartStartTime'] = np.bytes_(
                            sec.start_time.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))
                    else:
                        h5_file['Acquisition/Raw[0]/RawData'].\
                            attrs['PartStartTime'] = sec.start_time.strftime(
                                '%Y-%m-%dT%H:%M:%S.%f%z')
                    stime = sec.start_time.timestamp() * 1e6
                    DataTime = np.arange(
                        stime, stime + sec.nt / sec.fs, 1 / sec.fs)
                else:
                    h5_file['Acquisition/Raw[0]/RawData'].\
                        attrs['PartStartTime'] = np.bytes_(str(sec.start_time))
                    DataTime = sec.start_time + np.arange(0, sec.nt / sec.fs,
                                                        1 / sec.fs)
                _update_h5_dataset(h5_file, 'Acquisition/Raw[0]/',
                                   'RawDataTime', DataTime)
                h5_file['Acquisition/Raw[0]'].attrs['OutputDataRate'] = sec.fs
                h5_file['Acquisition'].attrs['SpatialSamplingInterval'] = sec.dx
                if hasattr(sec, 'gauge_length'):
                    h5_file['Acquisition'].attrs['GaugeLength'] = \
                        sec.gauge_length
            elif group == 'raw':
                _update_h5_dataset(h5_file, '/', 'raw', sec.data)
                DataTime = sec.start_time.timestamp() + \
                    np.arange(0, sec.nt / sec.fs, 1 / sec.fs)
                _update_h5_dataset(h5_file, '/', 'timestamp', DataTime)
            elif group == 'data': # https://ai4eps.github.io/homepage/ml4earth/seismic_event_format_das/
                _update_h5_dataset(h5_file, '/', 'data', sec.data)
                h5_file['data'].attrs['dx_m'] = sec.dx
                h5_file['data'].attrs['dt_s'] = 1 / sec.fs
                h5_file['data'].attrs['begin_time'] = \
                    datetime.strftime(sec.start_time, '%Y-%m-%dT%H:%M:%S.%f%z')
                h5_file['data'].attrs['unit'] = sec.data_type
            elif group == 'data_product':
                _update_h5_dataset(h5_file, 'data_product/', 'data', sec.data)
                h5_file.attrs['dt_computer'] = 1 / sec.fs
                h5_file.attrs['dx'] = sec.dx
                h5_file.attrs['gauge_length'] = sec.gauge_length
                DataTime = sec.start_time.timestamp() + \
                    np.arange(0, sec.nt / sec.fs, 1 / sec.fs)
                if h5_file.attrs['saving_start_gps_time'] > 0:
                    h5_file.attrs['file_start_gps_time'] = \
                        sec.start_time.timestamp()
                    _update_h5_dataset(h5_file, 'data_product/', 'gps_time',
                                       DataTime)
                    del h5_file['data_product/posix_time']
                else:
                    h5_file.attrs['file_start_computer_time'] = \
                        sec.start_time.timestamp()
                    _update_h5_dataset(h5_file, 'data_product/', 'posix_time',
                                       DataTime)
                    del h5_file['data_product/gps_time']
                h5_file.attrs['data_product'] = sec.data_type
            else:
                acquisition = list(h5_file[f'{group}/Source1/Zone1'].keys())[0]
                data = sec.data
                fs = int(sec.fs)
                d = len(h5_file[f'{group}/Source1/Zone1/{acquisition}'].shape)
                if d == 3:
                    mod = sec.nt % fs
                    if mod:
                        data = np.hstack((data, np.zeros((sec.nch, fs - mod))))
                    data = data.reshape((sec.nch, fs, sec.nt//fs)).T
                elif d == 2:
                    data = data.T
                _update_h5_dataset(h5_file, f'{group}/Source1/Zone1/',
                                   acquisition, data)

                h5_file[f'{group}/Source1/Zone1'].attrs['Spacing'][0] = sec.dx
                h5_file[f'{group}/Source1/Zone1'].attrs['FreqRes'] = \
                    np.bytes_(sec.fs)
                h5_file[f'{group}/Source1/Zone1'].attrs['SamplingRate'][0] = \
                    sec.fs
                h5_file[f'{group}/Source1/Zone1'].attrs['Extent'][0] = \
                    sec.start_channel
                h5_file[f'{group}/Source1/Zone1'].attrs['Origin'][0] = \
                    sec.start_distance
                h5_file[f'{group}/Source1/Zone1'].attrs['GaugeLength'][0] = \
                    sec.gauge_length
                DataTime = sec.start_time.timestamp() + \
                    np.arange(0, sec.nt / sec.fs, 1 / sec.fs)
                _update_h5_dataset(h5_file, f'{group}/Source1/',
                                   'time', DataTime.reshape((1, -1)))

    return None


def _write_segy(sec, fname, raw_fname=None):
    spec = segyio.spec()
    spec.samples = np.arange(sec.nt) / sec.fs * 1e3
    spec.tracecount = sec.nch
    if raw_fname is None:
        spec.format = 1
        with segyio.create(fname, spec) as new_file:
            new_file.header.length = sec.nch
            new_file.header.segy._filename = fname
            new_file.trace = sec.data.astype(np.float32)
    else:
        with segyio.open(raw_fname, ignore_geometry=True) as raw_file:
            spec.sorting = raw_file.sorting
            spec.format = raw_file.format
            raw_file.header.length = sec.nch
            raw_file.header.segy._filename = fname
            with segyio.create(fname, spec) as new_file:
                new_file.text[0] = raw_file.text[0]
                new_file.header = raw_file.header
                new_file.trace = sec.data.astype(raw_file.trace.dtype)

    warnings.warn('This data format doesn\'t include channel interval.')
    return None



# Purpose: Remove noise from data
# Author: Minzhe Hu, Zefeng Li
# Date: 2024.5.13
# Email: hmz2018@mail.ustc.edu.cn
import numpy as np
from copy import deepcopy
from scipy.ndimage import median_filter
from scipy.interpolate import interp1d
from daspy.basic_tools.preprocessing import padding
from daspy.advanced_tools.fdct import fdct_wrapping, ifdct_wrapping


def spike_removal(data, nch=50, nsp=5, thresh=10):
    """
    Use a median filter to remove high-strain spikes in the data. Modified from
    https://github.com/atterholt/curvelet-denoising/blob/main/MedianFilter.m

    :param data: numpy.ndarray. Data to remove spikes from.
    :param nch: int. Number of channels over which to compute the median.
    :param nsp: int. Number of sampling points over which to compute the median.
    :param thresh: Ratio threshold over the median over which a number is
        considered to be an outlier.
    :return: numpy.ndarray. Data with spikes removed.
    """
    absdata = np.abs(data)

    medians1 = median_filter(absdata, (nch, 1))
    medians = median_filter(medians1, (1, nsp))
    ratio = absdata / medians  # comparisons matrix

    # find the bad values and interpolate with their neighbors
    data_dn = data.copy()
    out_i, out_j = np.where(ratio > thresh)
    for j in set(out_j):
        bch = out_i[out_j == j]
        gch = list(set(range(len(data))) - set(bch))
        f = interp1d(gch, data[gch, j], bounds_error=False,
                     fill_value=(data[gch[0], j], data[gch[-1], j]))
        data_dn[bch, j] = f(bch)

    return data_dn


def common_mode_noise_removal(data, method='median'):
    """
    Remove common mode noise (sometimes called horizontal noise) from data.

    :param data: numpy.ndarray. Data to remove common mode noise.
    :param method:str. Method for extracting commmon mode noise. 'median' or
        'mean'
    :return: numpy.ndarray. Denoised data.
    """
    nch, nt = data.shape
    if method == 'median':
        common = np.median(data, 0)
    elif method == 'mean':
        common = np.mean(data, 0)

    xx = np.sum(common ** 2)
    data_dn = np.zeros((nch, nt))
    for i in range(nch):
        xc = np.sum(common * data[i])
        data_dn[i] = data[i] - xc / xx * common

    return data_dn


def _noise_level(data, finest=2, nbscales=None, nbangles=16, percentile=95):
    """
    Find threshold for curvelet denoising with noise record.

    :param data: numpy.ndarray. Noise data.
    :param nbscales: int. Number of scales including the coarsest wavelet level.
        Default set to ceil(log2(min(M,N)) - 3).
    :param nbangles: int. Number of angles at the 2nd coarsest level,
        minimum 8, must be a multiple of 4.
    :param percentile: number. The threshold is taken as this percentile of the
        curvelet coefficient of the noise record
    :return: 2-D list. Threshold for curvelet coefficients.
    """
    C = fdct_wrapping(data, is_real=True, finest=finest, nbscales=nbscales,
                      nbangles_coarse=nbangles)

    E_noise = []
    for s in range(len(C)):
        E_noise.append([])
        for w in range(len(C[s])):
            threshold = np.percentile(abs(C[s][w]), percentile)
            E_noise[s].append(threshold)

    return E_noise


def _knee_points(C, factor=0.2):
    """
    Find threshold for curvelet denoising without noise record.

    :param C: 2-D list of np.ndarray. Array of curvelet coefficients.
    :param factor: float. Multiplication factor from 0 to 1. Small factor
        corresponds to conservative strategy.
    :return: 2-D list. Threshold for curvelet coefficients.
    """
    E_knee = []
    for s in range(len(C)):
        E_knee.append([])
        for w in range(len(C[s])):
            F, x = np.histogram(abs(C[s][w]), density=True)
            x = (x[1:] + x[:-1]) / 2
            F = np.cumsum(F) / np.sum(F)
            slope = (x[-1] - x[0]) / (F[-1] - F[0])
            tiltedplot = x - (slope * F)
            idx = np.argmin(tiltedplot)
            E_knee[s].append(x[idx] * factor)

    return E_knee


def _velocity_bin(nbangles, fs, dx):
    v_bounds = np.zeros(nbangles // 4 + 1)
    half = nbangles // 8
    v_bounds[half] = fs * dx
    np.seterr(divide='ignore')
    for i in range(half):
        v_bounds[i] = i / half * fs * dx
        v_bounds[half + i + 1] = np.divide(fs * dx, 1 - (i + 1) / half)

    np.seterr(divide='warn')
    v_lows = list(range(half - 1, -1, -1)) + list(range(half * 2)) + \
        list(range(2 * half - 1, half - 1, -1))
    velocity = []
    for i in range(nbangles // 2):
        v_low = v_bounds[v_lows[i]]
        v_high = v_bounds[v_lows[i] + 1]
        velocity.append([v_low, v_high])
    velocity = np.array(velocity * 2)
    for i in range(half):
        velocity[i] = -1 * velocity[i][::-1]
        velocity[3 * half + i] = -1 * velocity[3 * half + i][::-1]
        velocity[4 * half + i] = -1 * velocity[4 * half + i][::-1]
        velocity[7 * half + i] = -1 * velocity[7 * half + i][::-1]
    return velocity


def _mask_factor(velocity, vmin, vmax, flag=0):
    if flag:
        if flag == -1:
            vmin = -vmax
            vmax = -vmin
    else:
        half = len(velocity) // 8
        for i in range(half):
            velocity[i] = -1 * velocity[i][::-1]
            velocity[3 * half + i] = -1 * velocity[3 * half + i][::-1]
            velocity[4 * half + i] = -1 * velocity[4 * half + i][::-1]
            velocity[7 * half + i] = -1 * velocity[7 * half + i][::-1]

    factors = np.zeros(len(velocity))
    for i, (v_low, v_high) in enumerate(velocity):
        v1 = max(v_low, vmin)
        v2 = min(v_high, vmax)
        if v1 < v2:
            if v_high == np.inf or v_low == -np.inf:
                factors[i] = 1
            else:
                factors[i] = np.divide(v2 - v1, v_high - v_low)

    return factors


def curvelet_denoising(data, choice=0, pad=0.3, noise=None, noise_perc=95,
                       knee_fac=0.2, soft_thresh=True, vmin=0, vmax=np.inf,
                       flag=0, dx=None, fs=None, mode='remove',
                       scale_begin=3, nbscales=None, nbangles=16, finest=2):
    """
    Use curevelet transform to filter stochastic or/and coherent noise.
    Modified from
    https://github.com/atterholt/curvelet-denoising/blob/main/CurveletDenoising.m
    {Atterholt et al., 2022 , Geophys. J. Int.}

    :param data: numpy.ndarray. Data to denoise.
    :param choice: int. 0 for Gaussian denoising using soft thresholding, 1 for
        velocity filtering using the standard FK methodology and 2 for both.
    :param pad: float or sequence of floats. Each float means padding percentage
        before FFT for corresponding dimension. If set to 0.1 will pad 5% before
        the beginning and after the end.
    :param noise: numpy.ndarray or daspy.Section. Noise record as reference.
    :param noise_perc: number. The threshold is taken as this percentile of the
        curvelet coefficient of the noise record. (only used when noise is
        specified)
    :param knee_fac: float. Multiplication factor from 0 to 1. Small factor
        corresponds to conservative strategy. (only used when noise is not
        specified)
    :param soft_thresh: bool. True for soft thresholding and False for hard
        thresholding.
    :param vmin, vmax: float. Velocity range in m/s.
    :param flag: -1 choose only negative apparent velocities, 0 choose both
        postive and negative apparent velocities, 1 choose only positive
        apparent velocities.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param mode: str. Only available when choice in (1,2). 'remove' for
        denoising, 'retain' for extraction, and 'decompose' for decomposition.
    :param scale_begin: int. The beginning scale to do coherent denoising.
    :param nbscales: int. Number of scales including the coarsest wavelet level.
        Default set to ceil(log2(min(M,N)) - 3).
    :param nbangles: int. Number of angles at the 2nd coarsest level,
        minimum 8, must be a multiple of 4.
    :param finest: int. Objects at the finest scale. 1 for curvelets, 2 for
        wavelets. Curvelets are more precise while wavelets are more efficient.
    :return: numpy.ndarray. Denoised data.
    """
    if pad is None or pad is False:
        pad = 0
    dn = np.round(np.array(pad) * data.shape).astype(int)
    data_pd = padding(data, dn)

    C = fdct_wrapping(data_pd, is_real=True, finest=finest, nbscales=nbscales,
                      nbangles_coarse=nbangles)

    # apply Gaussian denoising
    if choice in (0, 2):
        # define threshold
        if noise is None:
            E = _knee_points(C, factor=knee_fac)
        else:
            if not isinstance(noise, np.ndarray):
                noise = noise.data
            noise_pd = padding(noise,
                               np.array(data_pd.shape) - np.array(noise.shape))
            E = _noise_level(noise_pd, finest=finest, nbscales=nbscales,
                             nbangles=nbangles, percentile=noise_perc)
        for s in range(1, len(C)):
            for w in range(len(C[s])):
                # first do a hard threshold
                C[s][w] = C[s][w] * (abs(C[s][w]) > abs(E[s][w]))
                if soft_thresh:
                    # soften the existing coefficients
                    C[s][w] = np.sign(C[s][w]) * (abs(C[s][w]) - abs(E[s][w]))

    # apply velocity filtering
    if choice in (1, 2):
        if dx is None or fs is None:
            raise ValueError('Please set both dx and fs.')

        if mode == 'decompose':
            lst = list(range(scale_begin - 1)) 
            if finest == 2:
                lst.append(len(C) - 1)
            for s in lst:
                for w in range(len(C[s])):
                    C[s][w] /= 2
            C_rt = deepcopy(C)

        for s in range(scale_begin - 1, len(C) - finest + 1):
            nbangles = len(C[s])
            velocity = _velocity_bin(nbangles, fs, dx)
            factors = _mask_factor(velocity, vmin, vmax, flag=flag)
            for w in range(nbangles):
                if mode == 'retain':
                    C[s][w] *= factors[w]
                elif mode == 'remove':
                    C[s][w] *= 1 - factors[w]
                elif mode == 'decompose':
                    C[s][w] *= factors[w]
                    C_rt[s][w] *= 1 - factors[w]

    # perform the inverse curvelet transform
    data_dn = padding(ifdct_wrapping(C, is_real=True, size=data_pd.shape), dn,
                      reverse=True)
    
    if mode == 'decompose':
        data_n = padding(ifdct_wrapping(C_rt, is_real=True, size=data_pd.shape),
                         dn, reverse=True)
        return data_dn, data_n
    else:
        return data_dn


# Purpose: Convert strain rate data to velocity
# Author: Minzhe Hu
# Date: 2024.3.10
# Email: hmz2018@mail.ustc.edu.cn
import numpy as np
from numpy.fft import irfft2, ifftshift
from scipy.signal import hilbert
from daspy.basic_tools.freqattributes import next_pow_2, fk_transform
from daspy.basic_tools.preprocessing import padding, cosine_taper
from daspy.basic_tools.filter import bandpass
from daspy.advanced_tools.fdct import fdct_wrapping, ifdct_wrapping
from daspy.advanced_tools.denoising import _velocity_bin
from daspy.advanced_tools.decomposition import fk_fan_mask


def fk_rescaling(data, dx, fs, taper=(0.02, 0.05), pad='default', fmax=None,
                 kmin=(1 / 2000, 1 / 3000), vmax=(15000, 30000), edge=0.2,
                 turning=None, verbose=False):
    """
    Convert strain/strain rate to velocity/acceleration by fk rescaling.

    :param data: numpy.ndarray. Data to do fk rescaling.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param taper: float or sequence of floats. Each float means decimal
        percentage of Tukey taper for corresponding dimension (ranging from 0 to
        1). Default is 0.1 which tapers 5% from the beginning and 5% from the
        end.
    :param pad: Pad the data or not. It can be float or sequence of floats. Each
        float means padding percentage before FFT for corresponding dimension.
        If set to 0.1 will pad 5% before the beginning and after the end.
        'default' means pad both dimensions to next power of 2. None or False
        means don't pad data before or during Fast Fourier Transform.
    :param fmax, kmin, vmax: float or or sequence of 2 floats. Sequence of 2
        floats represents the start and end of taper. Setting these parameters
        can reduce artifacts.
    :param edge: float. The width of fan mask taper edge.
    :param turning: Sequence of int. Channel number of turning points.
    :param verbose: If True, return converted data, f-k spectrum, frequency
        sequence, wavenumber sequence and f-k mask.
    :return: Converted data and some variables in the process if verbose==True.
    """
    if turning is not None:
        data_vel = np.zeros_like(data)
        start_ch = [0, *turning]
        end_ch = [*turning, len(data)]
        for (s, e) in zip(start_ch, end_ch):
            data_vel[s:e] = fk_rescaling(data[s:e], dx, fs, taper=taper,
                                         pad=pad, fmax=fmax, kmin=kmin,
                                         vmax=vmax, edge=edge, verbose=False)
    else:
        data_tp = cosine_taper(data, taper)

        if pad == 'default':
            nch, nt = data.shape
            dn = (next_pow_2(nch) - nch, next_pow_2(nt) - nt)
            nfft = None
        elif pad is None or pad is False:
            dn = 0
            nfft = None
        else:
            dn = np.round(np.array(pad) * data.shape).astype(int)
            nfft = 'default'

        data_pd = padding(data_tp, dn)
        nch, nt = data_pd.shape

        fk, f, k = fk_transform(data_pd, dx, fs, taper=taper, nfft=nfft)

        ff = np.tile(f, (len(k), 1))
        kk = np.tile(k, (len(f), 1)).T
        vv = - np.divide(ff, kk, out=np.ones_like(ff) * 1e10, where=kk != 0)

        mask = fk_fan_mask(f, k, fmax=fmax, kmin=kmin, vmax=vmax, edge=edge) * vv
        mask[kk == 0] = 0

        data_vel = irfft2(ifftshift(fk * mask, axes=0)).real[:nch, :nt]
        data_vel = padding(data_vel, dn, reverse=True)

        if verbose:
            return data_vel, fk, f, k, mask
    return data_vel


def curvelet_conversion(data, dx, fs, pad=0.3, scale_begin=2, nbscales=None,
                        nbangles=16, turning=None):
    """
    Use curevelet transform to convert strain/strain rate to
    velocity/acceleration. {Yang et al. , 2023, Geophys. Res. Lett.}

    :param data: numpy.ndarray. Data to convert.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param pad: float or sequence of floats. Each float means padding percentage
        before FFT for corresponding dimension. If set to 0.1 will pad 5% before
        the beginning and after the end.
    :param scale_begin: int. The beginning scale to do conversion.
    :param nbscales: int. Number of scales including the coarsest wavelet level.
        Default set to ceil(log2(min(M,N)) - 3).
    :param nbangles: int. Number of angles at the 2nd coarsest level,
        minimum 8, must be a multiple of 4.
    :param turning: Sequence of int. Channel number of turning points.
    :return: numpy.ndarray. Converted data.
    """
    if turning is not None:
        data_vel = np.zeros_like(data)
        start_ch = [0, *turning]
        end_ch = [*turning, len(data)]
        for (s, e) in zip(start_ch, end_ch):
            data_vel[s:e] = curvelet_conversion(data[s:e], dx, fs, pad=pad,
                                                scale_begin=scale_begin,
                                                nbscales=nbscales,
                                                nbangles=nbangles, turning=None)
    else:
        if pad is None or pad is False:
            pad = 0
        dn = np.round(np.array(pad) * data.shape).astype(int)
        data_pd = padding(data, dn)

        C = fdct_wrapping(data_pd, is_real=True, finest=1, nbscales=nbscales,
                          nbangles_coarse=nbangles)

        # rescale with velocity
        np.seterr(divide='ignore')
        for s in range(0, scale_begin - 1):
            for w in range(len(C[s])):
                C[s][w] *= 0

        for s in range(scale_begin - 1, len(C)):
            nbangles = len(C[s])
            velocity = _velocity_bin(nbangles, fs, dx)
            factors = np.mean(velocity, axis=1)
            for w in range(nbangles):
                if abs(factors[w]) == np.inf:
                    factors[w] = abs(velocity[w]).min() * \
                        np.sign(velocity[w, 0]) * 2
                C[s][w] *= factors[w]

        data_vel = ifdct_wrapping(C, is_real=True, size=data_pd.shape)
        data_vel = padding(data_vel, dn, reverse=True)

    return data_vel


def slowness(g, dx, fs, slm, sls, swin=2):
    """
    Estimate the slowness time series by calculate semblance.
    {Lior et al., 2021, Solid Earth}

    :param g: 2-dimensional array. time series of adjacent channels used for
        estimating slowness
    :param dx: float. Spatical sampling rate (in m)
    :param fs: float. Sampling rate of records
    :param slm: float. Slowness x max
    :param sls: float. Slowness step
    :param swin: int. Slowness smooth window
    :return: Sequences of slowness and sembalence.
    """
    L = (len(g) - 1) // 2
    nt = len(g[0])
    h = np.imag(hilbert(g))
    grdpnt = round(slm / sls)
    sem = np.zeros((2 * grdpnt + 1, nt))
    gap = round(slm * dx * L * fs)

    h_ex = np.zeros((len(g), nt + 2 * gap))
    h_ex[:, gap:-gap] = h
    g_ex = np.zeros((len(g), nt + 2 * gap))
    g_ex[:, gap:-gap] = g

    for i in range(2 * grdpnt + 1):
        px = (i - grdpnt) * sls
        if abs(px) < 1e-5:
            continue
        gt = np.zeros(g.shape)
        ht = np.zeros(h.shape)
        for j in range(-L, L):
            shift = round(px * j * dx * fs)
            gt[j + L] = g_ex[j + L, gap + shift:gap + shift + nt]
            ht[j + L] = h_ex[j + L, gap + shift:gap + shift + nt]
        sem[i] = (np.sum(gt, axis=0)**2 + np.sum(ht, axis=0)**2) / \
            np.sum(gt**2 + ht**2, axis=0) / (2 * L + 1)
    p = (np.argmax(sem, axis=0) - grdpnt) * sls
    # smooth P
    for i in range(swin, nt - swin):
        win = p[i - swin:i + swin + 1]
        sign = np.sign(sum(np.sign(win)))
        win = [px for px in win if np.sign(px) == sign]
        p[i] = np.mean(win)

    return p, sem


def slant_stacking(data, dx, fs, L=None, slm=0.01,
                   sls=0.000125, frqlow=0.1, frqhigh=15, turning=None,
                   channel='all'):
    """
    Convert strain to velocity based on slant-stack.

    :param data: 2-dimensional array. Axis 0 is channel number and axis 1 is
        time series
    :param dx: float. Spatical sampling rate (in m)
    :param L: int. the number of adjacent channels over which slowness is
        estimated
    :param slm: float. Slowness x max
    :param sls: float. slowness step
    :param freqmin: Pass band low corner frequency.
    :param freqmax: Pass band high corner frequency.
    :param turning: Sequence of int. Channel number of turning points.
    :param channel: int or list or 'all'. convert a certain channel number /
        certain channel range / all channels.
    :return: Converted velocity data
    """
    if L is None:
        L = round(50 / dx)

    nch, nt = data.shape
    if isinstance(channel, str) and channel == 'all':
        channel = list(range(nch))
    elif isinstance(channel, int):
        channel = [channel]

    if turning is not None:
        data_vel = np.zeros((0, len(data[0])))
        start_ch = [0, *turning]
        end_ch = [*turning, len(data)]
        for (s, e) in zip(start_ch, end_ch):
            channel_seg = [ch-s for ch in range(s,e) if ch in channel]
            if len(channel_seg):
                d_vel = slant_stacking(data[s:e], dx, fs, L=L, slm=slm, sls=sls,
                                       frqlow=frqlow, frqhigh=frqhigh,
                                       turning=None, channel=channel_seg)
                data_vel = np.vstack((data_vel, d_vel))
    else:
        data_ex = padding(data, (2 * L, 0))
        swin = int(max((1 / frqhigh * fs) // 2, 1))
        data_vel = np.zeros((len(channel), nt))
        for i, ch in enumerate(channel):
            p, _ = slowness(data_ex[ch:ch + 2 * L + 1], dx, fs, slm, sls,
                            swin=swin)
            data_vel[i] = bandpass(data[ch] / p, fs=fs, freqmin=frqlow,
                                   freqmax=frqhigh)

    return data_vel



# Purpose: Analyze frequency attribute and transform in frequency domain
# Author: Minzhe Hu
# Date: 2024.6.8
# Email: hmz2018@mail.ustc.edu.cn
import numpy as np
from numpy.fft import rfft, rfft2, fftshift, fftfreq, rfftfreq
from scipy.signal import stft
from daspy.basic_tools.preprocessing import demeaning, detrending, cosine_taper


def next_pow_2(i):
    """
    Find the next power of two.

    :param i: float or int.
    :return: int. The next power of two for i.
    """
    buf = np.ceil(np.log2(i))
    return np.power(2, buf).astype(int)


def spectrum(data, fs, taper=0.05, nfft='default'):
    """
    Computes the spectrum of the given data.

    :param data: numpy.ndarray. Data to make spectrum of.
    :param fs: Sampling rate in Hz.
    :param taper: Decimal percentage of Tukey taper.
    :param nfft: Number of points for FFT. None = sampling points, 'default' =
        next power of 2 of sampling points.
    :return: Spectrum and frequency sequence.
    """
    if len(data.shape) == 1:
        data = data.reshape(1, len(data))
    elif len(data.shape) != 2:
        raise ValueError("Data should be 1-D or 2-D array")
    data = cosine_taper(data, (0, taper))

    if nfft == 'default':
        nfft = next_pow_2(len(data[0]))
    elif nfft is None:
        nfft = len(data[0])

    spec = rfft(data, n=nfft, axis=1)
    f = rfftfreq(nfft, d=1 / fs)

    return spec, f


def spectrogram(data, fs, nperseg=256, noverlap=None, nfft=None, detrend=False,
                boundary='zeros'):
    """
    Computes the spectrogram of the given data.

    :param data: 1-D or 2-D numpy.ndarray. Data to make spectrogram of.
    :param fs: Sampling rate in Hz.
    :param nperseg: int. Length of each segment.
    :param noverlap: int. Number of points to overlap between segments. If None,
        noverlap = nperseg // 2.
    :param nfft: int. Length of the FFT used. None = nperseg.
    :param detrend : str or bool. Specifies whether and how to detrend each
        segment.  'linear' or 'detrend' or True = detrend, 'constant' or
        'demean' = demean.
    :param boundary: str or None. Specifies whether the input signal is extended
        at both ends, and how to generate the new values, in order to center the
        first windowed segment on the first input point. This has the benefit of
        enabling reconstruction of the first input point when the employed
        window function starts at zero. Valid options are ['even', 'odd',
        'constant', 'zeros', None].
    :return: Spectrogram, frequency sequence and time sequence.
    """
    if detrend in [True, 'linear', 'detrend']:
        detrend = detrending
    elif detrend in ['constant', 'demean']:
        detrend = demeaning
    if data.ndim == 1:
        f, t, Zxx = stft(data, fs=fs, nperseg=nperseg, noverlap=noverlap,
                         nfft=nfft, detrend=detrend, boundary=boundary)
    elif len(data) == 1:
        f, t, Zxx = stft(data[0], fs=fs, nperseg=nperseg, noverlap=noverlap,
                         nfft=nfft, detrend=detrend, boundary=boundary)
    else:
        Zxx = []
        for d in data:
            f, t, Zxxi = stft(d, fs=fs, nperseg=nperseg, noverlap=noverlap,
                              nfft=nfft, detrend=detrend, boundary=boundary)
            Zxx.append(abs(Zxxi))
        Zxx = np.mean(np.array(Zxx), axis=0)

    return Zxx, f, t


def fk_transform(data, dx, fs, taper=(0, 0.05), nfft='default'):
    """
    Transform the data to the fk domain using 2-D Fourier transform method.

    :param data: numpy.ndarray. Data to do fk transform.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param taper: float or sequence of floats. Each float means decimal
        percentage of Tukey taper for corresponding dimension (ranging from 0 to
        1). Default is 0.1 which tapers 5% from the beginning and 5% from the
        end.
    :param nfft: Number of points for FFT. None means sampling points; 'default'
        means next power of 2 of sampling points, which makes result smoother.
    """
    nch, nt = data.shape
    data = cosine_taper(data, taper)
    if nfft == 'default':
        nfft = (next_pow_2(nch), next_pow_2(nt))
    elif not nfft:
        nfft = (nch, nt)

    fk = fftshift(rfft2(data, s=nfft), axes=0)
    f = rfftfreq(nfft[1], d=1. / fs)
    k = fftshift(fftfreq(nfft[0], d=dx))
    return fk, f, k



# Purpose: Module for reading DAS data.
# Author: Minzhe Hu
# Date: 2024.11.20
# Email: hmz2018@mail.ustc.edu.cn
# Partially modified from
# https://github.com/RobbinLuo/das-toolkit/blob/main/DasTools/DasPrep.py
import warnings
import json
import pickle
import numpy as np
import h5py
import segyio
from pathlib import Path
from nptdms import TdmsFile
from daspy.core.section import Section
from daspy.core.dasdatetime import DASDateTime, utc


def read(fname=None, output_type='section', ftype=None, headonly=False,
         **kwargs):
    """
    Read a .pkl/.pickle, .tdms, .h5/.hdf5, .segy/.sgy file.

    :param fname: str or pathlib.PosixPath. Path of DAS data file.
    :param output_type: str. 'Section' means return an instance of
        daspy.Section, 'array' means return numpy.array for data and a
        dictionary for metadata.
    :param ftype: None, str or function. None for automatic detection, or str to
        specify a type of 'pkl', 'pickle', 'tdms', 'h5', 'hdf5', 'segy', 'sgy',
        or 'npy', or a function for read data and metadata.
    :param headonly. bool. If True, only metadata will be read, the returned
        data will be an array of all zeros of the same size as the original
        data.
    :param ch1: int. The first channel required.
    :param ch2: int. The last channel required (not included).
    :param dch: int. Channel step.
    :return: An instance of daspy.Section, or numpy.array for data and a
        dictionary for metadata.
    """
    fun_map = {'pkl': _read_pkl, 'tdms': _read_tdms, 'h5': _read_h5,
               'sgy': _read_segy, 'npy': _read_npy}
    if fname is None:
        fname = Path(__file__).parent / 'example.pkl'
        ftype = 'pkl'
    if ftype is None:
        ftype = str(fname).split('.')[-1].lower()

    if callable(ftype):
        try:
            data, metadata = ftype(fname, headonly=headonly, **kwargs)
        except TypeError:
            data, metadata = ftype(fname)
    else:
        for rtp in [('pickle', 'pkl'), ('hdf5', 'h5'), ('segy', 'sgy')]:
            ftype = ftype.replace(*rtp)
        data, metadata = fun_map[ftype](fname, headonly=headonly, **kwargs)

    if output_type.lower() == 'section':
        metadata['source'] = Path(fname)
        metadata['source_type'] = ftype
        return Section(data.astype(float), **metadata)
    elif output_type.lower() == 'array':
        return data, metadata


def _read_pkl(fname, headonly=False, **kwargs):
    dch = kwargs.pop('dch', 1)
    with open(fname, 'rb') as f:
        pkl_data = pickle.load(f)
        if isinstance(pkl_data, np.ndarray):
            warnings.warn('This data format doesn\'t include channel interval'
                          'and sampling rate. Please set manually')
            if headonly:
                return np.zeros_like(pkl_data), {'dx': None, 'fs': None}
            else:
                ch1 = kwargs.pop('ch1', 0)
                ch2 = kwargs.pop('ch2', len(pkl_data))
                return pkl_data[ch1:ch2:dch], {'dx': None, 'fs': None}
        elif isinstance(pkl_data, dict):
            data = pkl_data.pop('data')
            if headonly:
                data = np.zeros_like(data)
            else:
                if 'ch1' in kwargs.keys() or 'ch2' in kwargs.keys():
                    if 'start_channel' in pkl_data.keys():
                        s_chn = pkl_data['start_channel']
                        print(f'Data is start with channel {s_chn}.')
                    else:
                        s_chn = 0
                    ch1 = kwargs.pop('ch1', s_chn)
                    ch2 = kwargs.pop('ch2', s_chn + len(data))
                    data = data[ch1 - s_chn:ch2 - s_chn, :]
                    pkl_data['start_channel'] = ch1
            return data, pkl_data
        else:
            raise TypeError('Unknown data type.')


def _read_h5_headers(group):
    headers = {}
    if len(group.attrs) != 0:
        headers['attrs'] = dict(group.attrs)
    if isinstance(group, h5py._hl.dataset.Dataset):
        return headers
    for key, value in group.items():
        try:
            gp_headers = _read_h5_headers(value)
        except AttributeError:
            headers[key] = value
        if len(gp_headers):
            headers[key] = gp_headers

    return headers


def _read_h5_starttime(h5_file):
    try:
        stime = h5_file['Acquisition/Raw[0]/RawData'].attrs['PartStartTime']
    except KeyError:
        try:
            stime = h5_file['Acquisition'].attrs['MeasurementStartTime']
        except KeyError:
            try:
                stime = h5_file['Acquisition/Raw[0]/RawDataTime/'][0]
            except KeyError:
                return 0
    if isinstance(stime, bytes):
        stime = stime.decode('ascii')

    if isinstance(stime, str):
        if len(stime) > 26:
            stime = DASDateTime.strptime(stime, '%Y-%m-%dT%H:%M:%S.%f%z')
        else:
            stime = DASDateTime.strptime(stime, '%Y-%m-%dT%H:%M:%S.%f').\
                astimezone(utc)
    else:
        stime = DASDateTime.fromtimestamp(stime / 1e6).astimezone(utc)

    return stime


def _read_h5(fname, headonly=False, **kwargs):
    with h5py.File(fname, 'r') as h5_file:
        dch = kwargs.pop('dch', 1)
        group = list(h5_file.keys())[0]
        if len(h5_file.keys()) >= 10: # ASN/OptoDAS https://github.com/ASN-Norway/simpleDAS
            ch1 = kwargs.pop('ch1', 0)
            if h5_file['header/dimensionNames'][0] == b'time':
                nch = h5_file['data'].shape[1]
                if headonly:
                    data = np.zeros_like(h5_file['data']).T
                else:
                    ch2 = kwargs.pop('ch2', nch)
                    data = h5_file['data'][:, ch1:ch2:dch].T
            elif h5_file['header/dimensionNames'][0] == b'distance':
                nch = h5_file['data'].shape[1]
                if headonly:
                    data = np.zeros_like(h5_file['data'])
                else:
                    ch2 = kwargs.pop('ch2', nch)
                    data = h5_file['data'][ch1:ch2:dch, :]
            dx = h5_file['header/dx'][()]
            start_time = DASDateTime.fromtimestamp(
                h5_file['header/time'][()]).utc()
            metadata = {'dx': dx * dch, 'fs': 1 / h5_file['header/dt'][()],
                        'start_time': start_time, 'start_channel': ch1,
                        'start_distance': ch1 * dx,
                        'scale': h5_file['header/dataScale'][()]}
            if h5_file['header/gaugeLength'][()] != np.nan:
                metadata['guage_length'] = h5_file['header/gaugeLength'][()]
        elif len(h5_file.keys()) == 5: # AP Sensing
            # read data
            nch = h5_file['strain'].shape[1]
            ch1 = kwargs.pop('ch1', 0)
            ch2 = kwargs.pop('ch2', nch)
            if headonly:
                data = np.zeros_like(h5_file['strain']).T
            else:
                data = h5_file['strain'][:, ch1:ch2:dch].T

            # read metadata
            dx = h5_file['spatialsampling'][()]
            metadata = {'fs': h5_file['RepetitionFrequency'][()],
                        'dx': dx * dch, 'start_channel': ch1,
                        'start_distance': ch1 * dx,
                        'gauge_length': h5_file.get('GaugeLength')[()]}
        elif set(h5_file.keys()) == {'Mapping', 'Acquisition'}: # Silixa/iDAS
            nch = h5_file['Acquisition/Raw[0]'].attrs['NumberOfLoci']
            ch1 = kwargs.pop('ch1', 0)
            ch2 = kwargs.pop('ch2', nch)
            if h5_file['Acquisition/Raw[0]/RawData/'].shape[0] == nch:
                if headonly:
                    data = np.zeros_like(h5_file['Acquisition/Raw[0]/RawData/'])
                else:
                    data = h5_file['Acquisition/Raw[0]/RawData/']\
                        [ch1:ch2:dch, :]
            else:
                if headonly:
                    data = np.zeros_like(
                        h5_file['Acquisition/Raw[0]/RawData/']).T
                else:
                    data = h5_file['Acquisition/Raw[0]/RawData/']\
                        [:, ch1:ch2:dch].T

            dx = np.mean(h5_file['Mapping/MeasuredSpatialResolution'])
            start_distance = h5_file['Acquisition/Custom/UserSettings'].\
                    attrs['StartDistance'] + ch1 * dx
            h5_file['Acquisition/Raw[0]/RawData'].attrs['PartStartTime']
            fs = h5_file['Acquisition/Raw[0]'].attrs['OutputDataRate']
            gauge_length = h5_file['Acquisition'].attrs['GaugeLength']
            scale = h5_file['Acquisition/Raw[0]'].attrs['AmpScaling']
            geometry = np.vstack((h5_file['Mapping/Lon'],
                                  h5_file['Mapping/Lat'])).T
            metadata = {'dx': dx * dch, 'fs': fs, 'start_channel': ch1,
                        'start_distance': ch1 * dx,
                        'gauge_length': gauge_length, 'geometry': geometry,
                        'scale': scale}
            metadata['start_time'] = _read_h5_starttime(h5_file)
        elif group == 'Acquisition':
            # OptaSens/ODH, Silixa/iDAS, Sintela/Onyx, Smart Sensing/ZD DAS
            # read data
            try:
                nch = h5_file['Acquisition'].attrs['NumberOfLoci']
            except KeyError:
                nch = len(h5_file['Acquisition/Raw[0]/RawData/'])
            ch1 = kwargs.pop('ch1', 0)
            ch2 = kwargs.pop('ch2', nch)
            if h5_file['Acquisition/Raw[0]/RawData/'].shape[0] == nch:
                if headonly:
                    data = np.zeros_like(h5_file['Acquisition/Raw[0]/RawData/'])
                else:
                    data = h5_file['Acquisition/Raw[0]/RawData/']\
                        [ch1:ch2:dch, :]
            else:
                if headonly:
                    data = np.zeros_like(
                        h5_file['Acquisition/Raw[0]/RawData/']).T
                else:
                    data = h5_file['Acquisition/Raw[0]/RawData/']\
                        [:, ch1:ch2:dch].T

            # read metadata
            try:
                fs = h5_file['Acquisition/Raw[0]'].attrs['OutputDataRate']
            except KeyError:
                time_arr = h5_file['Acquisition/Raw[0]/RawDataTime/']
                fs = 1 / (np.diff(time_arr).mean() / 1e6)

            dx = h5_file['Acquisition'].attrs['SpatialSamplingInterval']
            gauge_length = h5_file['Acquisition'].attrs['GaugeLength']
            metadata = {'dx': dx * dch, 'fs': fs, 'start_channel': ch1,
                        'start_distance': ch1 * dx,
                        'gauge_length': gauge_length}

            metadata['start_time'] = _read_h5_starttime(h5_file)
        elif group == 'raw':
            nch = len(h5_file['raw'])
            ch1 = kwargs.pop('ch1', 0)
            ch2 = kwargs.pop('ch2', nch)
            if headonly:
                data = np.zeros_like(h5_file['raw'])
            else:
                data = h5_file['raw'][ch1:ch2:dch, :]
            fs = round(1 / np.diff(h5_file['timestamp']).mean())
            start_time = DASDateTime.fromtimestamp(
                h5_file['timestamp'][0]).astimezone(utc)
            warnings.warn('This data format doesn\'t include channel interval. '
                          'Please set manually')
            metadata = {'dx': None, 'fs': fs, 'start_channel': ch1,
                        'start_time': start_time}
        elif group == 'data': # https://ai4eps.github.io/homepage/ml4earth/seismic_event_format_das/
            nch = h5_file['data'].shape[1]
            ch1 = kwargs.pop('ch1', 0)
            ch2 = kwargs.pop('ch2', nch)
            dch = kwargs.pop('dch', 1)
            if headonly:
                data = np.zeros_like(h5_file['raw_data'])
            else:
                data = h5_file['data'][ch1:ch2:dch, :]
            attr = h5_file['data'].attrs
            dx = attr['dx_m']
            metadata = {'dx': dx, 'fs': 1 / attr['dt_s'], 'start_channel': ch1,
                        'start_distance': ch1 * dx,
                        'start_time': DASDateTime.strptime(
                            attr['begin_time'], '%Y-%m-%dT%H:%M:%S.%f%z'),
                        'data_type': attr['unit']}
            if 'event_time' in attr.keys():
                try:
                    origin_time = DASDateTime.strptime(
                        attr['event_time'], '%Y-%m-%dT%H:%M:%S.%f%z')
                except ValueError:
                    origin_time = DASDateTime.strptime(
                        attr['event_time'], '%Y-%m-%dT%H:%M:%S.%f')
                metadata['origin_time'] = origin_time

        elif group == 'data_product':
            # read data
            nch = h5_file.attrs['nx']
            ch1 = kwargs.pop('ch1', 0)
            ch2 = kwargs.pop('ch2', nch)
            array_shape = h5_file['data_product/data'].shape
            if array_shape[0] == nch:
                if headonly:
                    data = np.zeros_like(h5_file['data_product/data'])
                else:
                    data = h5_file['data_product/data'][ch1:ch2:dch, :]
            else:
                if headonly:
                    data = np.zeros_like(h5_file['data_product/data']).T
                else:
                    data = h5_file['data_product/data'][:, ch1:ch2:dch].T

            # read metadata
            fs = 1 / h5_file.attrs['dt_computer']
            dx = h5_file.attrs['dx']
            gauge_length = h5_file.attrs['gauge_length']
            if h5_file.attrs['saving_start_gps_time'] > 0:
                start_time = DASDateTime.fromtimestamp(
                    h5_file.attrs['file_start_gps_time'])
            else:
                start_time = DASDateTime.fromtimestamp(
                    h5_file.attrs['file_start_computer_time'])
            data_type = h5_file.attrs['data_product']

            metadata = {'dx': dx * dch, 'fs': fs, 'start_channel': ch1,
                        'start_distance': ch1 * dx,
                        'start_time': start_time.astimezone(utc),
                        'gauge_length': gauge_length, 'data_type': data_type}
        else: # Febus
            acquisition = list(h5_file[f'{group}/Source1/Zone1'].keys())[0]
            # read data
            start_channel = int(h5_file[f'{group}/Source1/Zone1'].
                                attrs['Extent'][0])
            dataset = h5_file[f'{group}/Source1/Zone1/{acquisition}']
            nch = dataset.shape[-1]
            ch1 = kwargs.pop('ch1', start_channel)
            ch2 = kwargs.pop('ch2', start_channel + nch)
            if headonly:
                data = np.zeros_like(dataset).T.reshape((nch, -1))
            else:
                if len(dataset.shape) == 3: # Febus A1-R
                    data = dataset[:, :, ch1 - start_channel:ch2 - start_channel
                                   :dch].T.reshape(((ch2 - ch1) // dch, -1))
                elif len(dataset.shape) == 2: # Febus A1
                    data = dataset[:, ch1 - start_channel:ch2 - start_channel:
                                   dch].T
            # read metadata
            attrs = h5_file[f'{group}/Source1/Zone1'].attrs
            dx = attrs['Spacing'][0]
            try:
                fs = float(attrs['FreqRes'])
            except KeyError:
                try:
                    fs = (attrs['PulseRateFreq'][0] /
                          attrs['SamplingRes'][0]) / 1000
                except KeyError:
                    fs = attrs['SamplingRate'][0]
            start_distance = attrs['Origin'][0]
            time = h5_file[f'{group}/Source1/time']
            if len(time.shape) == 2: # Febus A1-R
                start_time = DASDateTime.fromtimestamp(time[0, 0]).\
                    astimezone(utc)
            elif len(time.shape) == 1: # Febus A1
                start_time = DASDateTime.fromtimestamp(time[0]).astimezone(utc)
            gauge_length = attrs['GaugeLength'][0]
            metadata = {'dx': dx * dch, 'fs': fs, 'start_channel': ch1,
                        'start_distance': start_distance +
                                            (ch1 - start_channel) * dx,
                        'start_time': start_time, 'gauge_length': gauge_length}

        metadata['headers'] = _read_h5_headers(h5_file)

    return data, metadata


def _read_tdms(fname, headonly=False, **kwargs):
    # https://nptdms.readthedocs.io/en/stable/quickstart.html
    with TdmsFile.read(fname) as tdms_file:
        group_name = [group.name for group in tdms_file.groups()]
        if 'Measurement' in group_name:
            key = 'Measurement'
        elif 'DAS' in group_name:
            key = 'DAS'
        else:
            key = group_name[0]

        headers = {**tdms_file.properties, **tdms_file[key].properties}
        nch = len(tdms_file[key])
        dch = kwargs.pop('dch', 1)
        # read data
        if nch > 1:
            start_channel = min(int(channel.name) for channel in
                                tdms_file[key].channels())
            ch1 = max(kwargs.pop('ch1', start_channel), start_channel)
            ch2 = min(kwargs.pop('ch2', start_channel + nch),
                      start_channel + nch)
            if headonly:
                nt = len(tdms_file[key][str(start_channel)])
                data = np.zeros((nch, nt))
            else:
                data = np.asarray([tdms_file[key][str(ch)]
                                   for ch in range(ch1, ch2, dch)])
        elif nch == 1:
            try:
                start_channel = int(headers['Initial Channel'])
            except KeyError:
                start_channel = 0

            ch1 = max(kwargs.pop('ch1', start_channel), start_channel)
            nch = int(headers['Total Channels'])
            ch2 = min(kwargs.pop('ch2', start_channel + nch),
                      start_channel + nch)
            if headonly:
                data = np.zeros(len(tdms_file[key].channels()[0])).\
                    reshape((nch, -1))
            else:
                data = np.asarray(tdms_file[key].channels()[0]).\
                    reshape((-1, nch)).T
                data = data[ch1 - start_channel:ch2 - start_channel:dch]

        # read metadata
        try:
            dx = headers['SpatialResolution[m]']
        except KeyError:
            try:
                dx = headers['Spatial Resolution']
            except KeyError:
                dx = None

        try:
            fs = headers['SamplingFrequency[Hz]']
        except KeyError:
            try:
                fs = 1 / headers['Time Base']
            except KeyError:
                fs = None

        try:
            start_distance = headers['Start Distance (m)'] + \
                dx * (ch1 - start_channel)
        except KeyError:
            start_distance = dx * ch1

        try:
            start_time = DASDateTime.strptime(headers['ISO8601 Timestamp'],
                                              '%Y-%m-%dT%H:%M:%S.%f%z')
        except ValueError:
            start_time = DASDateTime.strptime(headers['ISO8601 Timestamp'],
                                              '%Y-%m-%dT%H:%M:%S.%f')
        except KeyError:
            start_time = 0
            for key in ['GPSTimeStamp', 'CPUTimeStamp', 'Trigger Time']:
                if key in headers.keys():
                    if headers[key]:
                        start_time = DASDateTime.from_datetime(headers[key].
                                                               item())
                        break

        if dx is not None:
            dx *= dch
        metadata = {'dx': dx, 'fs': fs, 'start_channel': ch1,
                    'start_distance': start_distance, 'start_time': start_time,
                    'headers': headers}

        if 'GaugeLength' in headers.keys():
            metadata['gauge_length'] = headers['GaugeLength']

    return data, metadata


def _read_segy(fname, headonly=False, **kwargs):
    # https://github.com/equinor/segyio-notebooks/blob/master/notebooks/basic/02_segy_quicklook.ipynb
    with segyio.open(fname, ignore_geometry=True) as segy_file:
        nch = segy_file.tracecount
        ch1 = kwargs.pop('ch1', 0)
        ch2 = kwargs.pop('ch2', nch)
        dch = kwargs.pop('dch', 1)

        # read data
        if headonly:
            data = np.zeros_like(segy_file.trace.raw[:])
        else:
            data = segy_file.trace.raw[ch1:ch2:dch]

        # read metadata:
        fs = 1 / (segyio.tools.dt(segy_file) / 1e6)
        metadata = {'dx': None, 'fs': fs, 'start_channel': ch1}
        warnings.warn('This data format doesn\'t include channel interval.'
                      'Please set manually')

        return data, metadata


def _read_npy(fname, headonly=False, **kwargs):
    data = np.load(fname)
    if headonly:
        return np.zeros_like(data), {'dx': None, 'fs': None}
    else:
        ch1 = kwargs.pop('ch1', 0)
        ch2 = kwargs.pop('ch2', len(data))
        dch = kwargs.pop('dch', 1)
        warnings.warn('This data format doesn\'t include channel interval and '
                    'sampling rate. Please set manually')
        return data[ch1:ch2:dch], {'dx': None, 'fs': None}


def read_json(fname, output_type='dict'):
    """
    Read .json metadata file. See {Lai et al. , 2024, Seismol. Res. Lett.}

    :param fname: str or pathlib.PosixPath. Path of json file.
    :param output_type: str. 'dict' means return a dictionary, and 'Section'
        means return a empty daspy.Section instance with metadata.
    :return: A dictionary of metadata or an instance of daspy.Section without
        data.
    """
    with open(fname, 'r') as fcc_file:
        headers = json.load(fcc_file)
    if output_type.lower() == 'dict':
        return headers
    elif output_type.lower() in ['section', 'sec']:
        if len(headers['Overview']['Interrogator']) > 1:
            case_type = 'Multiple interrogators, single cable'
            sec_num = len(headers['Overview']['Interrogator'])
            sec = []
            for interrogator in headers['Overview']['Interrogator']:
                nch = interrogator['Acquisition'][0]['Attributes']['number_of_channels']
                data = np.zeros((nch, 0))
                dx = interrogator['Acquisition'][0]['Attributes']['spatial_sampling_interval']
                fs = interrogator['Acquisition'][0]['Attributes']['acquisition_sample_rate']
                gauge_length = interrogator['Acquisition'][0]['Attributes']['gauge_length']
                sec.append(Section(data, dx, fs, gauge_length=gauge_length,
                                   headers=headers))
        elif len(headers['Overview']['Interrogator'][0]['Acquisition']) > 1:
            case_type = 'Active survey'
            sec_num = len(
                headers['Overview']['Interrogator'][0]['Acquisition'])
            sec = []
            for acquisition in headers['Overview']['Interrogator'][0]['Acquisition']:
                nch = acquisition['Attributes']['number_of_channels']
                data = np.zeros((nch, 0))
                dx = acquisition['Attributes']['spatial_sampling_interval']
                fs = acquisition['Attributes']['acquisition_sample_rate']
                gauge_length = acquisition['Attributes']['gauge_length']
                sec.append(Section(data, dx, fs, gauge_length=gauge_length,
                                   headers=headers))
        else:
            sec_num = 1
            if len(headers['Overview']['Cable']) > 1:
                case_type = 'Single interrogators, multiple cable'
            else:
                env = headers['Overview']['Cable'][0]['Attributes']['cable_environment']
                if env == 'trench':
                    case_type = 'Direct buried'
                elif env == 'conduit':
                    case_type = 'Dark fiber'
                elif env in ['wireline', 'outside borehole casing']:
                    case_type = 'Borehole cable'
            nch = headers['Overview']['Interrogator'][0]['Acquisition'][0]['Attributes']['number_of_channels']
            dx = headers['Overview']['Interrogator'][0]['Acquisition'][0]['Attributes']['spatial_sampling_interval']
            fs = headers['Overview']['Interrogator'][0]['Acquisition'][0]['Attributes']['acquisition_sample_rate']
            gauge_length = headers['Overview']['Interrogator'][0]['Acquisition'][0]['Attributes']['gauge_length']
            data = np.zeros((nch, 0))
            sec = Section(data, dx, fs, gauge_length=gauge_length,
                          headers=headers)

        print(f'For case of {case_type}, create {sec_num} empty daspy.Section '
              'instance(s)')
        return sec



# Purpose: Filter the waveform
# Author: Minzhe Hu
# Date: 2024.10.16
# Email: hmz2018@mail.ustc.edu.cn
# Modified from https://docs.obspy.org/_modules/obspy/signal/filter.html
import warnings
import numpy as np
from scipy.signal import cheb2ord, cheby2, hilbert, iirfilter, zpk2sos, sosfilt


def bandpass(data, fs, freqmin, freqmax, corners=4, zi=None, zerophase=True):
    """
    Filter data from 'freqmin' to 'freqmax' using Butterworth bandpass filter of
    'corners' corners.

    :param data: numpy.ndarray. Data to filter.
    :param fs: Sampling rate in Hz.
    :param freqmin: Pass band low corner frequency.
    :param freqmax: Pass band high corner frequency.
    :param corners: Filter corners / order.
    :param zi : None, 0, or array_like. Initial conditions for the cascaded
        filter delays. It is a vector of shape (n_sections, nch, 2). Set to 0 to
        trigger a output of the final filter delay values.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered data. Only valid when zi is None.
    :return: Filtered data and the final filter delay values (if zi is not
        None).
    """
    if len(data.shape) == 1:
        data = data[np.newaxis, :]
    fe = 0.5 * fs
    low = freqmin / fe
    high = freqmax / fe
    # raise for some bad scenarios
    if high - 1.0 > -1e-6:
        msg = ('Selected high corner frequency ({}) of bandpass is at or ' +
               'above Nyquist ({}). Applying a high-pass instead.').format(
            freqmax, fe)
        warnings.warn(msg)
        return highpass(data, freq=freqmin, fs=fs, corners=corners,
                        zerophase=zerophase)
    if low > 1:
        msg = 'Selected low corner frequency is above Nyquist.'
        raise ValueError(msg)
    z, p, k = iirfilter(corners, [low, high], btype='band', ftype='butter',
                        output='zpk')
    sos = zpk2sos(z, p, k)
    if zi is None:
        data_flt = sosfilt(sos, data)
        if zerophase:
            data_flt = sosfilt(sos, data_flt[:, ::-1])[:, ::-1]
        return data_flt
    elif isinstance(zi, (int, float)):
        zi = np.ones((sos.shape[0], len(data), 2)) * zi
    
    data_flt, zf = sosfilt(sos, data, zi=zi)
    return data_flt, zf


def bandstop(data, fs, freqmin, freqmax, corners=4, zi=None, zerophase=False):
    """
    Filter data removing data between frequencies 'freqmin' and 'freqmax' using
    Butterworth bandstop filter of 'corners' corners.

    :param data: numpy.ndarray. Data to filter.
    :param fs: Sampling rate in Hz.
    :param freqmin: Stop band low corner frequency.
    :param freqmax: Stop band high corner frequency.
    :param corners: Filter corners / order.
    :param zi : None, 0, or array_like. Initial conditions for the cascaded
        filter delays. It is a vector of shape (n_sections, nch, 2). Set to 0 to
        trigger a output of the final filter delay values.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered data. Only valid when zi is None.
    :return: Filtered data and the final filter delay values (if zi is not
        None).
    """
    if len(data.shape) == 1:
        data = data[np.newaxis, :]
    fe = 0.5 * fs
    low = freqmin / fe
    high = freqmax / fe
    # raise for some bad scenarios
    if high > 1:
        high = 1.0
        msg = 'Selected high corner frequency is above Nyquist. Setting ' + \
              'Nyquist as high corner.'
        warnings.warn(msg)
    if low > 1:
        msg = 'Selected low corner frequency is above Nyquist.'
        raise ValueError(msg)
    z, p, k = iirfilter(corners, [low, high],
                        btype='bandstop', ftype='butter', output='zpk')
    sos = zpk2sos(z, p, k)
    if zi is None:
        data_flt = sosfilt(sos, data)
        if zerophase:
            data_flt = sosfilt(sos, data_flt[:, ::-1])[:, ::-1]
        return data_flt
    elif isinstance(zi, (int, float)):
        zi = np.ones((sos.shape[0], len(data), 2)) * zi
    
    data_flt, zf = sosfilt(sos, data, zi=zi)
    return data_flt, zf


def lowpass(data, fs, freq, corners=4, zi=None, zerophase=False):
    """
    Filter data removing data over certain frequency 'freq' using Butterworth
    lowpass filter of 'corners' corners.

    :param data: numpy.ndarray. Data to filter.
    :param fs: Sampling rate in Hz.
    :param freq: Filter corner frequency.
    :param corners: Filter corners / order.
    :param zi : None, 0, or array_like. Initial conditions for the cascaded
        filter delays. It is a vector of shape (n_sections, nch, 2). Set to 0 to
        trigger a output of the final filter delay values.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered data. Only valid when zi is None.
    :return: Filtered data and the final filter delay values (if zi is not
        None).
    """
    if len(data.shape) == 1:
        data = data[np.newaxis, :]
    fe = 0.5 * fs
    f = freq / fe
    # raise for some bad scenarios
    if f > 1:
        f = 1.0
        msg = 'Selected corner frequency is above Nyquist. Setting Nyquist ' + \
              'as high corner.'
        warnings.warn(msg)
    z, p, k = iirfilter(corners, f, btype='lowpass', ftype='butter',
                        output='zpk')
    sos = zpk2sos(z, p, k)
    if zi is None:
        data_flt = sosfilt(sos, data)
        if zerophase:
            data_flt = sosfilt(sos, data_flt[:, ::-1])[:, ::-1]
        return data_flt
    elif isinstance(zi, (int, float)):
        zi = np.ones((sos.shape[0], len(data), 2)) * zi
    
    data_flt, zf = sosfilt(sos, data, zi=zi)
    return data_flt, zf


def lowpass_cheby_2(data, fs, freq, maxorder=12, zi=None, ba=False,
                    freq_passband=False):
    """
    Filter data by passing data only below a certain frequency. The main purpose
    of this cheby2 filter is downsampling. This method will iteratively design a
    filter, whose pass band frequency is determined dynamically, such that the
    values above the stop band frequency are lower than -96dB.

    :param data: numpy.ndarray. Data to filter.
    :param fs: Sampling rate in Hz.
    :param freq: The frequency above which signals are attenuated with 95 dB.
    :param maxorder: Maximal order of the designed cheby2 filter.
    :param zi : None, 0, or array_like. Initial conditions for the cascaded
        filter delays. It is a vector of shape (n_sections, nch, 2). Set to 0 to
        trigger a output of the final filter delay values.
    :param ba: If True return only the filter coefficients (b, a) instead of
        filtering.
    :param freq_passband: If True return additionally to the filtered data, the
        iteratively determined pass band frequency.
    :return: Filtered data, the final filter delay values (if zi is not None)
        and the determined pass band frequency (if freq_passband is True).
    """
    if data.ndim == 1:
        data = data[np.newaxis, :]

    nyquist = fs * 0.5
    # rp - maximum ripple of passband, rs - attenuation of stopband
    rp, rs, order = 1, 96, 1e99
    ws = freq / nyquist  # stop band frequency
    wp = ws  # pass band frequency
    # raise for some bad scenarios
    if ws > 1:
        ws = 1.0
        warnings.warn('Selected corner frequency is above Nyquist. Setting '
                      'Nyquist as high corner.')
    while True:
        if order <= maxorder:
            break
        wp = wp * 0.99
        order, wn = cheb2ord(wp, ws, rp, rs, analog=0)
    if ba:
        return cheby2(order, rs, wn, btype='low', analog=0, output='ba')
    z, p, k = cheby2(order, rs, wn, btype='low', analog=0, output='zpk')
    sos = zpk2sos(z, p, k)
    if zi is None:
        data_flt = sosfilt(sos, data)
        if freq_passband:
            return data_flt, wp * nyquist
        return data_flt
    elif isinstance(zi, (int, float)):
        zi = np.ones((sos.shape[0], len(data), 2)) * zi
    
    data_flt, zf = sosfilt(sos, data, zi=zi)
    if freq_passband:
        return data_flt, zf, wp * nyquist
    return data_flt, zf


def highpass(data, fs, freq, corners=4, zi=None, zerophase=False):
    """
    Filter data removing data below certain frequency 'freq' using Butterworth
    highpass filter of 'corners' corners.

    :param data: numpy.ndarray. Data to filter.
    :param fs: Sampling rate in Hz.
    :param freq: Filter corner frequency.
    :param corners: Filter corners / order.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered data. Only valid when zi is None.
    :return: Filtered data and the final filter delay values (if zi is not
        None).
    """
    if len(data.shape) == 1:
        data = data[np.newaxis, :]
    fe = 0.5 * fs
    f = freq / fe
    # raise for some bad scenarios
    if f > 1:
        msg = 'Selected corner frequency is above Nyquist.'
        raise ValueError(msg)
    z, p, k = iirfilter(corners, f, btype='highpass', ftype='butter',
                        output='zpk')
    sos = zpk2sos(z, p, k)
    if zi is None:
        data_flt = sosfilt(sos, data)
        if zerophase:
            data_flt = sosfilt(sos, data_flt[:, ::-1])[:, ::-1]
        return data_flt
    elif isinstance(zi, (int, float)):
        zi = np.ones((sos.shape[0], len(data), 2)) * zi
    
    data_flt, zf = sosfilt(sos, data, zi=zi)
    return data_flt, zf

def envelope(data):
    """
    Computes the envelope of the given data. The envelope is determined by
    adding the squared amplitudes of the data and it's Hilbert-Transform and
    then taking the square-root. The envelope at the start/end should not be
    taken too seriously.

    :param data: numpy.ndarray. Data to make envelope of.
    :return: Envelope of input data.
    """
    return abs(hilbert(data, axis=-1))



# Purpose: Some preprocess methods
# Author: Minzhe Hu
# Date: 2025.3.10
# Email: hmz2018@mail.ustc.edu.cn
import numpy as np
from scipy.signal import detrend
from scipy.signal.windows import tukey
from daspy.basic_tools.filter import lowpass_cheby_2


def phase2strain(data, lam, e, n, gl):
    """
    Convert the optical phase shift in radians to strain.

    :param data: numpy.ndarray. Data to convert.
    :param lam: float. Operational optical wavelength in vacuum.
    :param e: float. photo-slastic scaling factor for logitudinal strain in
        isotropic material.
    :param n: float. Refractive index of the sensing fiber.
    :paran gl: float. Gauge length.
    :return: Strain data.
    """
    return data * (lam * 1e-9) / (e * 4 * np.pi * n * gl)


def normalization(data, method='z-score'):
    """
    Normalize for each individual channel using Z-score method.

    :param data: numpy.ndarray. Data to normalize.
    :param method: str. Method for normalization, should be one of 'max',
        'z-score', 'MAD' or 'one-bit'.
    :return: Normalized data.
    """
    if data.ndim == 1:
        data = data.reshape(1, len(data))
    elif data.ndim != 2:
        raise ValueError("Data should be 1-D or 2-D array")

    if method.lower() == 'max':
        amp = np.max(abs(data), 1, keepdims=True)
        amp[amp == 0] = amp[amp > 0].min()
        return data / amp
    elif method.lower() == 'z-score':
        mean = np.mean(data, axis=1, keepdims=True)
        std = np.std(data, axis=1, keepdims=True)
        std[std == 0] = std[std > 0].min()
        return (data - mean) / std
    elif method.lower() == 'mad':
        median = np.median(data, axis=1, keepdims=True)
        mad = np.median(abs(data - median), axis=1, keepdims=True)
        mad[mad == 0] = mad[mad > 0].min()
        return (data - median) / mad
    elif method.lower() == 'one-bit':
        return np.sign(data)


def demeaning(data):
    """
    Demean signal by subtracted mean of each channel.

    :param data: numpy.ndarray. Data to demean.
    :return: Detrended data.
    """
    return detrend(data, type='constant')


def detrending(data):
    """
    Detrend signal by subtracted a linear least-squares fit to data.

    :param data: numpy.ndarray. Data to detrend.
    :return: Detrended data.
    """
    return detrend(data, type='linear')


def stacking(data: np.ndarray, N: int, step: int = None, average: bool = True):
    """
    Stack several channels to increase the signal-noise ratio(SNR).

    :param data: numpy.ndarray. Data to stack.
    :param N: int. N adjacent channels stacked into 1.
    :param step: int. Interval of data stacking.
    :param average: bool. True for calculating the average.
    :return: Stacked data.
    """
    if N == 1:
        return data
    if step is None:
        step = N
    nch, nt = data.shape
    begin = np.arange(0, nch - N + 1, step)
    end = begin + N
    nx1 = len(begin)
    data_stacked = np.zeros((nx1, nt))
    for i in range(nx1):
        data_stacked[i, :] = np.sum(data[begin[i]:end[i], :], axis=0)
    if average:
        data_stacked /= N
    return data_stacked


def cosine_taper(data, p=0.1, side='both'):
    """
    Taper using Tukey window.

    :param data: numpy.ndarray. Data to taper.
    :param p: float or sequence of floats. Each float means decimal percentage
        of Tukey taper for corresponding dimension (ranging from 0 to 1).
        Default is 0.1 which tapers 5% from the beginning and 5% from the end.
        If only one float is given, it only do for time dimension.
    :param side: str. 'both', 'left', or 'right'.
    :return: Tapered data.
    """
    if data.ndim == 1:
        data = data.reshape(1, -1)
    nch, nt = data.shape
    if not isinstance(p, (tuple, list, np.ndarray)):
        win = tukey(nt, p)
        if side == 'left':
            win[round(nch/2):] = 1
        elif side == 'right':
            win[:round(len(win)/2)] = 1
        return data * np.tile(win, (nch, 1))
    else:
        if p[0] > 0:
            data = data * np.tile(tukey(nch, p[0]), (nt, 1)).T
        return cosine_taper(data, p[1], side=side)


def downsampling(data, xint=None, tint=None, stack=True, lowpass_filter=True):
    """
    Downsample DAS data.

    :param data: numpy.ndarray. Data to downsample can be 1-D or 2-D.
    :param xint: int. Spatial downsampling factor.
    :param tint: int. Time downsampling factor.
    :param lowpass_filter: bool. Lowpass cheby2 filter before time downsampling
        or not.
    :return: Downsampled data.
    """
    data_ds = data.copy()
    if xint and xint > 1:
        if stack:
            data_ds = stacking(data, xint)
        else:
            data_ds = data_ds[::xint]
    if tint and tint > 1:
        if lowpass_filter:
            data_ds = lowpass_cheby_2(data_ds, 1, 1 / 2 / tint)
        if len(data_ds.shape) == 1:
            data_ds = data_ds[::tint]
        else:
            data_ds = data_ds[:, ::tint]
    return data_ds


def trimming(data, dx=None, fs=None, xmin=0, xmax=None, tmin=0, tmax=None,
             mode=0):
    """
    Cut data to given start and end distance/channel or time/sampling points.

    :param data: numpy.ndarray. Data to trim can be 1-D or 2-D.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param xmin, xmax, tmin, tmax: Boundary for trimming.
    :param mode: 0 means the unit of boundary is channel number and sampling
        points; 1 means the unit of boundary is meters and seconds.
    :return: Trimmed data.
    """
    nch, nt = data.shape
    if mode == 0:
        if xmax is None:
            xmax = nch
        if tmax is None:
            tmax = nt
    elif mode == 1:
        xmin = round(xmin / dx)
        xmax = (round(xmax / dx), nch)[xmax is None]
        tmin = round(tmin * fs)
        tmax = (round(tmax * fs), nt)[tmax is None]

    return data[xmin:xmax, tmin:tmax]


def padding(data, dn, reverse=False):
    """
    Pad DAS data with 0.

    :param data: numpy.ndarray. 2D DAS data to pad.
    :param dn: int or sequence of ints. Number of points to pad for both
        dimensions.
    :param reverse: bool. Set True to reverse the operation.
    :return: Padded data.
    """
    nch, nt = data.shape
    if isinstance(dn, int):
        dn = (dn, dn)

    pad = (dn[0] // 2, dn[0] - dn[0] // 2, dn[1] // 2, dn[1] - dn[1] // 2)
    if reverse:
        return data[pad[0]:nch - pad[1], pad[2]:nt - pad[3]]
    else:
        data_pd = np.zeros((nch + dn[0], nt + dn[1]))
        data_pd[pad[0]:nch + pad[0], pad[2]:nt + pad[2]] = data
        return data_pd


def time_integration(data, fs, c=0):
    """
    Integrate DAS data in time.

    :param data: numpy.ndarray. 2D DAS data.
    :param fs: Sampling rate in Hz.
    :param c: float. A constant added to the result.
    :return: Integrated data.
    """
    return np.cumsum(data, axis=1) / fs + c


def time_differential(data, fs, prepend=0):
    """
    Differentiate DAS data in time.

    :param data: numpy.ndarray. 2D DAS data.
    :param fs: Sampling rate in Hz.
    :param prepend: 'mean' or values to prepend to `data` along axis prior to
        performing the difference. 
    :return: Differentiated data.
    """
    if prepend == 'mean':
        prepend = np.mean(data, axis=1).reshape((-1, 1))
    return np.diff(data, axis=1, prepend=prepend) * fs


def distance_integration(data, dx, c=0):
    """
    Integrate DAS data in distance.

    :param data: numpy.ndarray. 2D DAS data.
    :param dx: Channel interval in m.
    :param c: float. A constant added to the result.
    :return: Integrated data.
    """
    return np.cumsum(data, axis=1) * dx + c


# Purpose: Fast Discrete Curvelet Transform
# Author: Minzhe Hu
# Date: 2024.4.11
# Email: hmz2018@mail.ustc.edu.cn
# Modified from
# http://www.curvelet.org/download-secure.php?file=CurveLab-2.1.3.tar.gz
# (matlab version)
import numpy as np
from numpy.fft import fftshift, ifftshift, fft2, ifft2


def _round(x):
    return np.round(x).astype(int)


def _floor(x):
    return np.floor(x).astype(int)


def _ceil(x):
    return np.ceil(x).astype(int)


def fdct_wrapping_window(x):
    """
    Creates the two halves of a C**inf compactly supported window.

    :param x: vector or matrix of abscissae, the relevant ones from 0 to 1.
    :return: vector or matrix containing samples of the left, resp. right half
        of the window.
    """

    # Initialize the variables
    wr = np.zeros_like(x)
    wl = np.zeros_like(x)

    # Set values close to zero to zero
    x[np.abs(x) < 2**-52] = 0

    # Calculate wr and wl
    wr[(x > 0) & (x < 1)] = np.exp(
        1 - 1. / (1 - np.exp(1 - 1. / x[(x > 0) & (x < 1)])))
    wr[x <= 0] = 1
    wl[(x > 0) & (x < 1)] = np.exp(
        1 - 1. / (1 - np.exp(1 - 1. / (1 - x[(x > 0) & (x < 1)]))))
    wl[x >= 1] = 1

    # Normalize wr and wl
    normalization = np.sqrt(wl**2 + wr**2)
    wr = wr / normalization
    wl = wl / normalization

    return wl, wr


def fdct_wrapping(x, is_real=False, finest=2,
                  nbscales=None, nbangles_coarse=16):
    """
    Fast Discrete Curvelet Transform via wedge wrapping.

    :param x: np.array. M-by-N matrix.
    :param is_real: bool. Type of the transform, False for complex-valued
        curvelets and True for real-valued curvelets.
    :param finest: int. Chooses one of two possibilities for the coefficients at
        the finest level: 1 for curvelets and 2 for wavelets.
    :param nbscales: int. Number of scales including the coarsest wavele
        level. Default set to ceil(log2(min(M,N)) - 3).
    :param nbangles_coarse: int. Number of angles at the 2nd coarsest level,
        minimum 8, must be a multiple of 4.
    :return: 2-D list of np.ndarray. Array of curvelet coefficients.
        C[j][l][k1,k2] is the coefficient at scale j(from finest to coarsest
        scale), angle l(starts at the top-left corner and increases clockwise),
        position k1, k2(size varies with j and l). If is_real is 1, there are
        two types of curvelets, 'cosine' and 'sine'. For a given scale j, the
        'cosine' coefficients are stored in the first two quadrants (low values
        of l), the 'sine' coefficients in the last two quadrants (high values of
        l).
    """
    X = fftshift(fft2(ifftshift(x))) / np.sqrt(x.size)
    N1, N2 = X.shape
    if nbscales is None:
        nbscales = _ceil(np.log2(min(N1, N2)) - 3)

    # Initialization: data structure
    nbangles = [1] + [nbangles_coarse * 2 ** ((nbscales - i) // 2)
                      for i in range(nbscales, 1, -1)]
    if finest == 2:
        nbangles[-1] = 1

    C = []
    for j in range(nbscales):
        C.append([None] * nbangles[j])

    # Loop: pyramidal scale decomposition
    M1 = N1 / 3
    M2 = N2 / 3

    if finest == 1:
        # Initialization: smooth periodic extension of high frequencies
        bigN1 = 2 * _floor(2 * M1) + 1
        bigN2 = 2 * _floor(2 * M2) + 1
        equiv_index_1 = (_floor(N1 / 2) - _floor(2 * M1) +
                         np.arange(bigN1)) % N1
        equiv_index_2 = (_floor(N2 / 2) - _floor(2 * M2) +
                         np.arange(bigN2)) % N2
        X = X[np.ix_(equiv_index_1, equiv_index_2)]

        window_length_1 = _floor(2 * M1) - _floor(M1) - (N1 % 3 == 0)
        window_length_2 = _floor(2 * M2) - _floor(M2) - (N2 % 3 == 0)
        coord_1 = np.linspace(0, 1, window_length_1)
        coord_2 = np.linspace(0, 1, window_length_2)
        wl_1, wr_1 = fdct_wrapping_window(coord_1)
        wl_2, wr_2 = fdct_wrapping_window(coord_2)

        lowpass_1 = np.concatenate((wl_1, np.ones(2 * _floor(M1) + 1), wr_1))
        if N1 % 3 == 0:
            lowpass_1 = np.concatenate(([0], lowpass_1, [0]))

        lowpass_2 = np.concatenate((wl_2, np.ones(2 * _floor(M2) + 1), wr_2))
        if N2 % 3 == 0:
            lowpass_2 = np.concatenate(([0], lowpass_2, [0]))

        lowpass = np.outer(lowpass_1, lowpass_2)
        Xlow = X * lowpass
        scales = np.arange(nbscales, 1, -1)

    else:
        M1 /= 2
        M2 /= 2

        window_length_1 = _floor(2 * M1) - _floor(M1)
        window_length_2 = _floor(2 * M2) - _floor(M2)
        coord_1 = np.linspace(0, 1, window_length_1)
        coord_2 = np.linspace(0, 1, window_length_2)
        wl_1, wr_1 = fdct_wrapping_window(coord_1)
        wl_2, wr_2 = fdct_wrapping_window(coord_2)

        lowpass_1 = np.concatenate((wl_1, np.ones(2 * _floor(M1) + 1), wr_1))
        lowpass_2 = np.concatenate((wl_2, np.ones(2 * _floor(M2) + 1), wr_2))
        lowpass = np.outer(lowpass_1, lowpass_2)
        hipass = np.sqrt(1 - lowpass ** 2)

        Xlow_index_1 = np.arange(-_floor(2 * M1),
                                 _floor(2 * M1) + 1) + _ceil((N1 + 1) / 2) - 1
        Xlow_index_2 = np.arange(-_floor(2 * M2),
                                 _floor(2 * M2) + 1) + _ceil((N2 + 1) / 2) - 1
        Xlow = X[np.ix_(Xlow_index_1, Xlow_index_2)] * lowpass
        Xhi = X.copy()
        Xhi[np.ix_(Xlow_index_1, Xlow_index_2)] *= hipass

        C[nbscales - 1][0] = fftshift(ifft2(ifftshift(Xhi))
                                      ) * np.sqrt(Xhi.size)
        if is_real:
            C[nbscales - 1][0] = C[nbscales - 1][0].real

        scales = np.arange(nbscales - 1, 1, -1)
    for j in scales - 1:
        M1 /= 2
        M2 /= 2
        window_length_1 = _floor(2 * M1) - _floor(M1)
        window_length_2 = _floor(2 * M2) - _floor(M2)
        coord_1 = np.linspace(0, 1, window_length_1)
        coord_2 = np.linspace(0, 1, window_length_2)
        wl_1, wr_1 = fdct_wrapping_window(coord_1)
        wl_2, wr_2 = fdct_wrapping_window(coord_2)

        lowpass_1 = np.concatenate((wl_1, np.ones(2 * _floor(M1) + 1), wr_1))
        lowpass_2 = np.concatenate((wl_2, np.ones(2 * _floor(M2) + 1), wr_2))
        lowpass = np.outer(lowpass_1, lowpass_2)
        hipass = np.sqrt(1 - lowpass ** 2)

        Xhi = Xlow.copy()
        Xlow_index_1 = np.arange(-_floor(2 * M1),
                                 _floor(2 * M1) + 1) + _floor(4 * M1)
        Xlow_index_2 = np.arange(-_floor(2 * M2),
                                 _floor(2 * M2) + 1) + _floor(4 * M2)
        Xlow = Xlow[np.ix_(Xlow_index_1, Xlow_index_2)]
        Xhi[np.ix_(Xlow_index_1, Xlow_index_2)] = Xlow * hipass
        Xlow *= lowpass

        # Loop: angular decomposition
        l = -1
        nbquadrants = 2 + 2 * (not is_real)
        nbangles_perquad = nbangles[j] // 4
        for quadrant in range(1, nbquadrants + 1):
            M_horiz = (M1, M2)[quadrant % 2]
            M_vert = (M2, M1)[quadrant % 2]
            wedge_ticks_left = _round(
                np.linspace(
                    0,
                    1,
                    nbangles_perquad +
                    1) *
                _floor(
                    4 *
                    M_horiz) +
                1)
            wedge_ticks_right = 2 * _floor(4 * M_horiz) + 2 - wedge_ticks_left
            if nbangles_perquad % 2:
                wedge_ticks = np.concatenate(
                    (wedge_ticks_left, wedge_ticks_right[::-1]))
            else:
                wedge_ticks = np.concatenate(
                    (wedge_ticks_left, wedge_ticks_right[-2::-1]))

            wedge_endpoints = wedge_ticks[1:-1:2]
            wedge_midpoints = (wedge_endpoints[:-1] + wedge_endpoints[1:]) / 2
            # Left corner wedge
            l += 1
            first_wedge_endpoint_vert = _round(
                _floor(4 * M_vert) / nbangles_perquad + 1)
            length_corner_wedge = _floor(4 * M_vert) - _floor(M_vert) + \
                _ceil(first_wedge_endpoint_vert / 4)
            Y_corner = np.arange(length_corner_wedge) + 1
            XX, YY = np.meshgrid(
                np.arange(2 * _floor(4 * M_horiz) + 1) + 1, Y_corner)
            width_wedge = wedge_endpoints[1] + wedge_endpoints[0] - 1
            slope_wedge = (_floor(4 * M_horiz) + 1 -
                           wedge_endpoints[0]) / _floor(4 * M_vert)
            left_line = _round(
                2 - wedge_endpoints[0] + slope_wedge * (Y_corner - 1))
            wrapped_data = np.zeros(
                (length_corner_wedge, width_wedge), dtype=complex)
            wrapped_XX = np.zeros(
                (length_corner_wedge, width_wedge), dtype=int)
            wrapped_YY = np.zeros(
                (length_corner_wedge, width_wedge), dtype=int)
            first_row = _floor(4 * M_vert) + 2 - \
                _ceil((length_corner_wedge + 1) / 2) + \
                (length_corner_wedge + 1) % 2 * (quadrant - 2 == quadrant % 2)
            first_col = _floor(4 * M_horiz) + 2 - _ceil((width_wedge + 1) / 2) \
                + (width_wedge + 1) % 2 * (quadrant - 3 == (quadrant - 3) % 2)
            for row in Y_corner - 1:
                cols = left_line[row] + \
                    (np.arange(width_wedge) - (left_line[row] - first_col)) \
                    % width_wedge
                admissible_cols = _round(0.5 * (cols + 1 + abs(cols - 1))) - 1
                new_row = (row - first_row + 1) % length_corner_wedge
                wrapped_data[new_row, :] = Xhi[row,
                                               admissible_cols] * (cols > 0)
                wrapped_XX[new_row, :] = XX[row, admissible_cols]
                wrapped_YY[new_row, :] = YY[row, admissible_cols]
            slope_wedge_right = (_floor(4 * M_horiz) + 1 -
                                 wedge_midpoints[0]) / _floor(4 * M_vert)
            mid_line_right = wedge_midpoints[0] + \
                slope_wedge_right * (wrapped_YY - 1)
            coord_right = 0.5 + _floor(4 * M_vert) / \
                (wedge_endpoints[1] - wedge_endpoints[0]) * \
                (wrapped_XX - mid_line_right) / \
                (_floor(4 * M_vert) + 1 - wrapped_YY)
            C2 = 1 / (1 / (2 * (_floor(4 * M_horiz)) / (wedge_endpoints[0] -
                1) - 1) + 1 / (2 * (_floor(4 * M_vert)) / (
                first_wedge_endpoint_vert - 1) - 1))
            C1 = C2 / (2 * (_floor(4 * M_vert)) /
                       (first_wedge_endpoint_vert - 1) - 1)
            wrapped_XX[(wrapped_XX - 1) / _floor(4 * M_horiz) +
                       (wrapped_YY - 1) / _floor(4 * M_vert) == 2] += 1
            coord_corner = C1 + C2 * ((wrapped_XX - 1) / _floor(4 * M_horiz) -
                (wrapped_YY - 1) / _floor(4 * M_vert)) / (2 -
                ((wrapped_XX - 1) / _floor(4 * M_horiz) + (wrapped_YY - 1) /
                _floor(4 * M_vert)))
            wl_left, _ = fdct_wrapping_window(coord_corner)
            _, wr_right = fdct_wrapping_window(coord_right)
            wrapped_data = wrapped_data * wl_left * wr_right
            if not is_real:
                wrapped_data = np.rot90(wrapped_data, -(quadrant - 1))
                C[j][l] = fftshift(ifft2(ifftshift(wrapped_data))) * \
                    np.sqrt(wrapped_data.size)
            else:
                wrapped_data = np.rot90(wrapped_data, -(quadrant - 1))
                x = fftshift(ifft2(ifftshift(wrapped_data))) * \
                    np.sqrt(wrapped_data.size)
                C[j][l] = np.sqrt(2) * x.real
                C[j][l + nbangles[j] // 2] = np.sqrt(2) * x.imag

            # Regular wedges
            length_wedge = _floor(4 * M_vert) - _floor(M_vert)
            Y = np.arange(length_wedge) + 1
            first_row = _floor(4 * M_vert) + 2 - _ceil((length_wedge + 1) / 2) \
                + (length_wedge + 1) % 2 * (quadrant - 2 == quadrant % 2)
            for subl in range(1, nbangles_perquad - 1):
                l += 1
                width_wedge = wedge_endpoints[subl +
                                              1] - wedge_endpoints[subl - 1] + 1
                slope_wedge = ((_floor(4 * M_horiz) + 1) -
                               wedge_endpoints[subl]) / _floor(4 * M_vert)
                left_line = _round(
                    wedge_endpoints[subl - 1] + slope_wedge * (Y - 1))
                wrapped_data = np.zeros(
                    (length_wedge, width_wedge), dtype=complex)
                wrapped_XX = np.zeros((length_wedge, width_wedge), dtype=int)
                wrapped_YY = np.zeros((length_wedge, width_wedge), dtype=int)
                first_col = _floor(4 * M_horiz) + 2 - \
                    _ceil((width_wedge + 1) / 2) + \
                    (width_wedge + 1) % 2 * (quadrant - 3 == (quadrant - 3) % 2)
                for row in Y - 1:
                    cols = left_line[row] + (np.arange(width_wedge) -
                        (left_line[row] - first_col)) % width_wedge - 1
                    new_row = (row - first_row + 1) % length_wedge
                    wrapped_data[new_row, :] = Xhi[row, cols]
                    wrapped_XX[new_row, :] = XX[row, cols]
                    wrapped_YY[new_row, :] = YY[row, cols]
                slope_wedge_left = ((_floor(4 * M_horiz) + 1) -
                    wedge_midpoints[subl - 1]) / _floor(4 * M_vert)
                mid_line_left = wedge_midpoints[subl - 1] + \
                    slope_wedge_left * (wrapped_YY - 1)
                coord_left = 0.5 + _floor(4 * M_vert) / \
                    (wedge_endpoints[subl] - wedge_endpoints[subl - 1]) * \
                    (wrapped_XX - mid_line_left) / \
                    (_floor(4 * M_vert) + 1 - wrapped_YY)
                slope_wedge_right = ((_floor(4 * M_horiz) + 1) -
                                     wedge_midpoints[subl]) / _floor(4 * M_vert)
                mid_line_right = wedge_midpoints[subl] + \
                    slope_wedge_right * (wrapped_YY - 1)
                coord_right = 0.5 + _floor(4 * M_vert) / \
                    (wedge_endpoints[subl + 1] - wedge_endpoints[subl]) * \
                    (wrapped_XX - mid_line_right) / \
                    (_floor(4 * M_vert) + 1 - wrapped_YY)

                wl_left, _ = fdct_wrapping_window(coord_left)
                _, wr_right = fdct_wrapping_window(coord_right)
                wrapped_data = wrapped_data * wl_left * wr_right
                if not is_real:
                    wrapped_data = np.rot90(wrapped_data, -(quadrant - 1))
                    C[j][l] = fftshift(ifft2(ifftshift(wrapped_data))) * \
                        np.sqrt(wrapped_data.size)
                else:
                    wrapped_data = np.rot90(wrapped_data, -(quadrant - 1))
                    x = fftshift(ifft2(ifftshift(wrapped_data))) * \
                        np.sqrt(wrapped_data.size)
                    C[j][l] = np.sqrt(2) * x.real
                    C[j][l + nbangles[j] // 2] = np.sqrt(2) * x.imag

            # Right corner wedge
            l += 1
            width_wedge = 4 * _floor(4 * M_horiz) + 3 - \
                wedge_endpoints[-1] - wedge_endpoints[-2]
            slope_wedge = ((_floor(4 * M_horiz) + 1) -
                           wedge_endpoints[-1]) / _floor(4 * M_vert)
            left_line = _round(
                wedge_endpoints[-2] + slope_wedge * (Y_corner - 1))
            wrapped_data = np.zeros((length_corner_wedge, width_wedge),
                                    dtype=complex)
            wrapped_XX = np.zeros((length_corner_wedge, width_wedge), dtype=int)
            wrapped_YY = np.zeros((length_corner_wedge, width_wedge), dtype=int)
            first_row = _floor(4 * M_vert) + 2 - \
                _ceil((length_corner_wedge + 1) / 2) + \
                (length_corner_wedge + 1) % 2 * (quadrant - 2 == quadrant % 2)
            first_col = _floor(4 * M_horiz) + 2 - _ceil((width_wedge + 1) / 2) + \
                (width_wedge + 1) % 2 * (quadrant - 3 == (quadrant - 3) % 2)
            for row in Y_corner - 1:
                cols = left_line[row] + (np.arange(width_wedge) -
                    (left_line[row] - first_col)) % width_wedge
                admissible_cols = _round(0.5 * (cols + 2 * _floor(4 * M_horiz)
                    + 1 - np.abs(cols - (2 * _floor(4 * M_horiz) + 1)))) - 1
                new_row = (row - first_row + 1) % length_corner_wedge
                wrapped_data[new_row, :] = Xhi[row, admissible_cols] * \
                    (cols <= (2 * _floor(4 * M_horiz) + 1))
                wrapped_XX[new_row, :] = XX[row, admissible_cols]
                wrapped_YY[new_row, :] = YY[row, admissible_cols]

            slope_wedge_left = ((_floor(4 * M_horiz) + 1) -
                                wedge_midpoints[-1]) / _floor(4 * M_vert)
            mid_line_left = wedge_midpoints[-1] + \
                slope_wedge_left * (wrapped_YY - 1)
            coord_left = 0.5 + _floor(4 * M_vert) / \
                (wedge_endpoints[-1] - wedge_endpoints[-2]) * \
                (wrapped_XX - mid_line_left) / \
                (_floor(4 * M_vert) + 1 - wrapped_YY)
            C2 = -1 / (2 * (_floor(4 * M_horiz)) / (wedge_endpoints[-1] - 1) -
                       1 + 1 / (2 * (_floor(4 * M_vert)) /
                                (first_wedge_endpoint_vert - 1) - 1))
            C1 = -C2 * (2 * (_floor(4 * M_horiz)) /
                        (wedge_endpoints[-1] - 1) - 1)
            wrapped_XX[(wrapped_XX - 1) / _floor(4 * M_horiz) ==
                       (wrapped_YY - 1) / _floor(4 * M_vert)] -= 1
            coord_corner = C1 + C2 * (2 - ((wrapped_XX - 1) /
                _floor(4 * M_horiz) + (wrapped_YY - 1) / _floor(4 * M_vert))) \
                / ((wrapped_XX - 1) / _floor(4 * M_horiz) - (wrapped_YY - 1) /
                _floor(4 * M_vert))
            wl_left, _ = fdct_wrapping_window(coord_left)
            _, wr_right = fdct_wrapping_window(coord_corner)
            wrapped_data = wrapped_data * wl_left * wr_right
            if not is_real:
                wrapped_data = np.rot90(wrapped_data, -(quadrant - 1))
                C[j][l] = fftshift(ifft2(ifftshift(wrapped_data))
                                   ) * np.sqrt(wrapped_data.size)
            else:
                wrapped_data = np.rot90(wrapped_data, -(quadrant - 1))
                x = fftshift(ifft2(ifftshift(wrapped_data))) * \
                    np.sqrt(wrapped_data.size)
                C[j][l] = np.sqrt(2) * x.real
                C[j][l + nbangles[j] // 2] = np.sqrt(2) * x.imag

            if quadrant < nbquadrants:
                Xhi = np.rot90(Xhi)
    # Coarsest wavelet level
    C[0][0] = fftshift(ifft2(ifftshift(Xlow))) * np.sqrt(Xlow.size)
    if is_real:
        C[0][0] = C[0][0].real

    return C


def ifdct_wrapping(C, is_real=False, size=None):
    """
    Inverse Fast Discrete Curvelet Transform via wedge wrapping. This is in fact
    the adjoint, also the pseudo-inverse

    :param C: 2-D list of np.ndarray. Array of curvelet coefficients.
    :param is_real: bool. Type of the transform, False for complex-valued
        curvelets and True for real-valued curvelets.
    :param size: tuple of ints. Size of the image to be recovered (not necessary
        if finest = 2)
    :return: 2-D np.ndarray.
    """
    nbscales = len(C)
    nbangles_coarse = len(C[1])
    nbangles = [1] + [nbangles_coarse * 2 ** ((nbscales - i) // 2)
                      for i in range(nbscales, 1, -1)]
    if len(C[-1]) == 1:
        finest = 2
        nbangles[nbscales - 1] = 1
    else:
        finest = 1

    if size is None:
        if finest == 1:
            raise ValueError("Require output size.")
        else:
            N1, N2 = C[-1][0].shape
    else:
        N1, N2 = size

    M1 = N1 / 3
    M2 = N2 / 3

    if finest == 1:
        # Initialization: preparing the lowpass filter at finest scale
        window_length_1 = _floor(2 * M1) - _floor(M1) - (N1 % 3 == 0)
        window_length_2 = _floor(2 * M2) - _floor(M2) - (N2 % 3 == 0)
        coord_1 = np.linspace(0, 1, window_length_1)
        coord_2 = np.linspace(0, 1, window_length_2)
        wl_1, wr_1 = fdct_wrapping_window(coord_1)
        wl_2, wr_2 = fdct_wrapping_window(coord_2)

        lowpass_1 = np.concatenate((wl_1, np.ones(2 * _floor(M1) + 1), wr_1))
        if N1 % 3 == 0:
            lowpass_1 = np.concatenate(([0], lowpass_1, [0]))

        lowpass_2 = np.concatenate((wl_2, np.ones(2 * _floor(M2) + 1), wr_2))
        if N2 % 3 == 0:
            lowpass_2 = np.concatenate(([0], lowpass_2, [0]))

        lowpass = np.outer(lowpass_1, lowpass_2)
        scales = np.arange(nbscales, 1, -1)
    else:
        M1 /= 2
        M2 /= 2

        window_length_1 = _floor(2 * M1) - _floor(M1)
        window_length_2 = _floor(2 * M2) - _floor(M2)
        coord_1 = np.linspace(0, 1, window_length_1)
        coord_2 = np.linspace(0, 1, window_length_2)
        wl_1, wr_1 = fdct_wrapping_window(coord_1)
        wl_2, wr_2 = fdct_wrapping_window(coord_2)

        lowpass_1 = np.concatenate((wl_1, np.ones(2 * _floor(M1) + 1), wr_1))
        lowpass_2 = np.concatenate((wl_2, np.ones(2 * _floor(M2) + 1), wr_2))
        lowpass = np.outer(lowpass_1, lowpass_2)
        hipass_finest = np.sqrt(1 - lowpass ** 2)

        scales = np.arange(nbscales - 1, 1, -1)

    bigN1 = 2 * _floor(2 * M1) + 1
    bigN2 = 2 * _floor(2 * M2) + 1
    X = np.zeros((bigN1, bigN2), dtype=complex)

    # Loop: pyramidal reconstruction

    Xj_topleft_1 = 1
    Xj_topleft_2 = 1
    for j in scales - 1:
        M1 /= 2
        M2 /= 2

        window_length_1 = _floor(2 * M1) - _floor(M1)
        window_length_2 = _floor(2 * M2) - _floor(M2)
        coord_1 = np.linspace(0, 1, window_length_1)
        coord_2 = np.linspace(0, 1, window_length_2)
        wl_1, wr_1 = fdct_wrapping_window(coord_1)
        wl_2, wr_2 = fdct_wrapping_window(coord_2)

        lowpass_1 = np.concatenate((wl_1, np.ones(2 * _floor(M1) + 1), wr_1))
        lowpass_2 = np.concatenate((wl_2, np.ones(2 * _floor(M2) + 1), wr_2))
        lowpass_next = np.outer(lowpass_1, lowpass_2)
        hipass = np.sqrt(1 - lowpass_next ** 2)
        Xj = np.zeros((2 * _floor(4 * M1) + 1, 2 * _floor(4 * M2) + 1),
                      dtype=complex)

        # Loop: angles
        l = -1
        nbquadrants = 2 + 2 * (not is_real)
        nbangles_perquad = nbangles[j] // 4
        for quadrant in range(1, nbquadrants + 1):
            M_horiz = (M1, M2)[quadrant % 2]
            M_vert = (M2, M1)[quadrant % 2]
            wedge_ticks_left = _round(np.linspace(0, 1, nbangles_perquad + 1) *
                                      _floor(4 * M_horiz) + 1)
            wedge_ticks_right = 2 * _floor(4 * M_horiz) + 2 - wedge_ticks_left
            if nbangles_perquad % 2:
                wedge_ticks = np.concatenate(
                    (wedge_ticks_left, wedge_ticks_right[::-1]))
            else:
                wedge_ticks = np.concatenate(
                    (wedge_ticks_left, wedge_ticks_right[-2::-1]))
            wedge_endpoints = wedge_ticks[1:-1:2]
            wedge_midpoints = (wedge_endpoints[:-1] + wedge_endpoints[1:]) / 2

            # Left corner wedge
            l += 1
            first_wedge_endpoint_vert = _round(_floor(4 * M_vert) /
                                               nbangles_perquad + 1)
            length_corner_wedge = _floor(4 * M_vert) - _floor(M_vert) + \
                _ceil(first_wedge_endpoint_vert / 4)
            Y_corner = np.arange(length_corner_wedge) + 1
            [XX, YY] = np.meshgrid(np.arange(1, 2 * _floor(4 * M_horiz) + 2),
                                   Y_corner)
            width_wedge = wedge_endpoints[1] + wedge_endpoints[0] - 1
            slope_wedge = (_floor(4 * M_horiz) + 1 -
                           wedge_endpoints[0]) / _floor(4 * M_vert)
            left_line = _round(2 - wedge_endpoints[0] +
                               slope_wedge * (Y_corner - 1))
            wrapped_XX = np.zeros((length_corner_wedge, width_wedge), dtype=int)
            wrapped_YY = np.zeros((length_corner_wedge, width_wedge), dtype=int)
            first_row = _floor(4 * M_vert) + \
                2 - _ceil((length_corner_wedge + 1) / 2) + \
                (length_corner_wedge + 1) % 2 * (quadrant - 2 == quadrant % 2)
            first_col = _floor(4 * M_horiz) + 2 - _ceil((width_wedge + 1) / 2) \
                + (width_wedge + 1) % 2 * (quadrant - 3 == (quadrant - 3) % 2)
            for row in Y_corner - 1:
                cols = left_line[row] + (np.arange(width_wedge) -
                    (left_line[row] - first_col)) % width_wedge
                new_row = (row - first_row + 1) % length_corner_wedge
                admissible_cols = _round(0.5 * (cols + 1 + abs(cols - 1))) - 1
                wrapped_XX[new_row, :] = XX[row, admissible_cols]
                wrapped_YY[new_row, :] = YY[row, admissible_cols]

            slope_wedge_right = (_floor(4 * M_horiz) + 1 - wedge_midpoints[0]) \
                / _floor(4 * M_vert)
            mid_line_right = wedge_midpoints[0] + \
                slope_wedge_right * (wrapped_YY - 1)
            coord_right = 0.5 + _floor(4 * M_vert) / (wedge_endpoints[1] -
                wedge_endpoints[0]) * (wrapped_XX - mid_line_right) / \
                (_floor(4 * M_vert) + 1 - wrapped_YY)
            C2 = 1 / (1 / (2 * (_floor(4 * M_horiz)) /
                (wedge_endpoints[0] - 1) - 1) + 1 / (2 * (_floor(4 * M_vert))
                / (first_wedge_endpoint_vert - 1) - 1))
            C1 = C2 / (2 * (_floor(4 * M_vert)) /
                       (first_wedge_endpoint_vert - 1) - 1)
            wrapped_XX[(wrapped_XX - 1) / _floor(4 * M_horiz) +
                       (wrapped_YY - 1) / _floor(4 * M_vert) == 2] += 1
            coord_corner = C1 + C2 * ((wrapped_XX - 1) / _floor(4 * M_horiz) -
                (wrapped_YY - 1) / _floor(4 * M_vert)) / (2 - ((wrapped_XX - 1)
                / _floor(4 * M_horiz) + (wrapped_YY - 1) / _floor(4 * M_vert)))
            wl_left, _ = fdct_wrapping_window(coord_corner)
            _, wr_right = fdct_wrapping_window(coord_right)

            if not is_real:
                wrapped_data = fftshift(fft2(ifftshift(C[j][l]))) / \
                    np.sqrt(C[j][l].size)
                wrapped_data = np.rot90(wrapped_data, quadrant - 1)
            else:
                x = C[j][l] + 1j * C[j][l + nbangles[j] // 2]
                wrapped_data = fftshift(fft2(ifftshift(x))) / \
                    np.sqrt(x.size * 2)
                wrapped_data = np.rot90(wrapped_data, quadrant - 1)

            wrapped_data = wrapped_data * wl_left * wr_right
            # Unwrapping data
            for row in Y_corner - 1:
                cols = left_line[row] + (np.arange(width_wedge) -
                    (left_line[row] - first_col)) % width_wedge
                admissible_cols = _round(0.5 * (cols + 1 + abs(cols - 1))) - 1
                new_row = (row - first_row + 1) % length_corner_wedge
                Xj[row, admissible_cols] += wrapped_data[new_row, :]
                # We use the following property: in an assignment A(B) = C where
                # B and C are vectors, if some value x repeats in B, then the
                # last occurrence of x is the one corresponding to the eventual
                # assignment.

            # Regular wedges
            length_wedge = _floor(4 * M_vert) - _floor(M_vert)
            Y = np.arange(length_wedge) + 1
            first_row = _floor(4 * M_vert) + 2 - _ceil((length_wedge + 1) / 2) \
                + (length_wedge + 1) % 2 * (quadrant - 2 == quadrant % 2)
            for subl in range(1, nbangles_perquad - 1):
                l += 1
                width_wedge = wedge_endpoints[subl + 1] - \
                    wedge_endpoints[subl - 1] + 1
                slope_wedge = ((_floor(4 * M_horiz) + 1) -
                               wedge_endpoints[subl]) / _floor(4 * M_vert)
                left_line = _round(wedge_endpoints[subl - 1] +
                                   slope_wedge * (Y - 1))
                wrapped_XX = np.zeros((length_wedge, width_wedge), dtype=int)
                wrapped_YY = np.zeros((length_wedge, width_wedge), dtype=int)
                first_col = _floor(4 * M_horiz) + 2 - \
                    _ceil((width_wedge + 1) / 2) + \
                    (width_wedge + 1) % 2 * (quadrant - 3 == (quadrant - 3) % 2)
                for row in Y - 1:
                    cols = left_line[row] + (np.arange(width_wedge) -
                        (left_line[row] - first_col)) % width_wedge - 1
                    new_row = (row - first_row + 1) % length_wedge
                    wrapped_XX[new_row, :] = XX[row, cols]
                    wrapped_YY[new_row, :] = YY[row, cols]

                slope_wedge_left = ((_floor(4 * M_horiz) + 1) -
                    wedge_midpoints[subl - 1]) / _floor(4 * M_vert)
                mid_line_left = wedge_midpoints[subl - 1] + \
                    slope_wedge_left * (wrapped_YY - 1)
                coord_left = 0.5 + _floor(4 * M_vert) / (wedge_endpoints[subl]
                    - wedge_endpoints[subl - 1]) * \
                    (wrapped_XX - mid_line_left) / \
                    (_floor(4 * M_vert) + 1 - wrapped_YY)
                slope_wedge_right = ((_floor(4 * M_horiz) + 1) -
                                     wedge_midpoints[subl]) / _floor(4 * M_vert)
                mid_line_right = wedge_midpoints[subl] + \
                    slope_wedge_right * (wrapped_YY - 1)
                coord_right = 0.5 + _floor(4 * M_vert) / \
                    (wedge_endpoints[subl + 1] - wedge_endpoints[subl]) * \
                    (wrapped_XX - mid_line_right) / \
                    (_floor(4 * M_vert) + 1 - wrapped_YY)
                wl_left, _ = fdct_wrapping_window(coord_left)
                _, wr_right = fdct_wrapping_window(coord_right)
                if not is_real:
                    wrapped_data = fftshift(fft2(ifftshift(C[j][l]))) / \
                        np.sqrt(C[j][l].size)
                    wrapped_data = np.rot90(wrapped_data, quadrant - 1)
                else:
                    x = C[j][l] + 1j * C[j][l + nbangles[j] // 2]
                    wrapped_data = fftshift(
                        fft2(ifftshift(x))) / np.sqrt(x.size * 2)
                    wrapped_data = np.rot90(wrapped_data, quadrant - 1)

                wrapped_data = wrapped_data * wl_left * wr_right

                # Unwrapping data
                for row in Y - 1:
                    cols = left_line[row] + (np.arange(width_wedge) -
                        (left_line[row] - first_col)) % width_wedge - 1
                    new_row = (row + 1 - first_row) % length_wedge
                    Xj[row, cols] += wrapped_data[new_row, :]

            # Right corner wedge
            l += 1
            width_wedge = 4 * _floor(4 * M_horiz) + 3 - \
                wedge_endpoints[-1] - wedge_endpoints[-2]
            slope_wedge = ((_floor(4 * M_horiz) + 1) -
                           wedge_endpoints[-1]) / _floor(4 * M_vert)
            left_line = _round(
                wedge_endpoints[-2] + slope_wedge * (Y_corner - 1))
            wrapped_XX = np.zeros(
                (length_corner_wedge, width_wedge), dtype=int)
            wrapped_YY = np.zeros(
                (length_corner_wedge, width_wedge), dtype=int)
            first_row = _floor(4 * M_vert) + 2 - \
                _ceil((length_corner_wedge + 1) / 2) + \
                (length_corner_wedge + 1) % 2 * (quadrant - 2 == quadrant % 2)
            first_col = _floor(4 * M_horiz) + 2 - _ceil((width_wedge + 1) / 2) \
                + (width_wedge + 1) % 2 * (quadrant - 3 == (quadrant - 3) % 2)
            for row in Y_corner - 1:
                cols = left_line[row] + (np.arange(width_wedge) -
                    (left_line[row] - first_col)) % width_wedge
                admissible_cols = _round(0.5 * (cols + 2 * _floor(4 * M_horiz)
                    + 1 - np.abs(cols - (2 * _floor(4 * M_horiz) + 1)))) - 1
                new_row = (row - first_row + 1) % length_corner_wedge
                wrapped_XX[new_row, :] = XX[row, admissible_cols]
                wrapped_YY[new_row, :] = YY[row, admissible_cols]

            slope_wedge_left = ((_floor(4 * M_horiz) + 1) -
                                wedge_midpoints[-1]) / _floor(4 * M_vert)
            mid_line_left = wedge_midpoints[-1] + \
                slope_wedge_left * (wrapped_YY - 1)
            coord_left = 0.5 + _floor(4 * M_vert) / \
                (wedge_endpoints[-1] - wedge_endpoints[-2]) * \
                (wrapped_XX - mid_line_left) / \
                (_floor(4 * M_vert) + 1 - wrapped_YY)
            C2 = -1 / (2 * (_floor(4 * M_horiz)) / (wedge_endpoints[-1] - 1)
                       - 1 + 1 / (2 * (_floor(4 * M_vert)) /
                                  (first_wedge_endpoint_vert - 1) - 1))
            C1 = -C2 * (2 * (_floor(4 * M_horiz)) /
                        (wedge_endpoints[-1] - 1) - 1)

            wrapped_XX[(wrapped_XX - 1) / _floor(4 * M_horiz) ==
                       (wrapped_YY - 1) / _floor(4 * M_vert)] -= 1
            coord_corner = C1 + C2 * (2 - ((wrapped_XX - 1) /
                _floor(4 * M_horiz) + (wrapped_YY - 1) / _floor(4 * M_vert))) \
                / ((wrapped_XX - 1) / _floor(4 * M_horiz) - (wrapped_YY - 1) /
                _floor(4 * M_vert))
            wl_left, _ = fdct_wrapping_window(coord_left)
            _, wr_right = fdct_wrapping_window(coord_corner)

            if not is_real:
                wrapped_data = fftshift(
                    fft2(ifftshift(C[j][l]))) / np.sqrt(C[j][l].size)
                wrapped_data = np.rot90(wrapped_data, quadrant - 1)
            else:
                x = C[j][l] + 1j * C[j][l + nbangles[j] // 2]
                wrapped_data = fftshift(
                    fft2(ifftshift(x))) / np.sqrt(x.size * 2)
                wrapped_data = np.rot90(wrapped_data, quadrant - 1)

            wrapped_data = wrapped_data * wl_left * wr_right

            # Unwrapping data
            for row in Y_corner - 1:
                cols = left_line[row] + (np.arange(width_wedge) -
                    (left_line[row] - first_col)) % width_wedge
                admissible_cols = _round(1 / 2 * (cols + 2 * _floor(4 * M_horiz)
                    + 1 - abs(cols - (2 * _floor(4 * M_horiz) + 1)))) - 1
                new_row = (row + 1 - first_row) % length_corner_wedge
                Xj[row, np.flip(admissible_cols)] += wrapped_data[new_row, ::-1]
                # We use the following property: in an assignment A[B] = C where
                # B and C are vectors, if some value x repeats in B, then the
                # last occurrence of x is the one corresponding to the eventual
                # assignment.

            Xj = np.rot90(Xj)

        Xj *= lowpass
        Xj_index1 = np.arange(-_floor(2 * M1),
                              _floor(2 * M1) + 1) + _floor(4 * M1)
        Xj_index2 = np.arange(-_floor(2 * M2),
                              _floor(2 * M2) + 1) + _floor(4 * M2)

        Xj[np.ix_(Xj_index1, Xj_index2)] *= hipass

        loc_1 = Xj_topleft_1 + np.arange(2 * _floor(4 * M1) + 1) - 1
        loc_2 = Xj_topleft_2 + np.arange(2 * _floor(4 * M2) + 1) - 1
        X[np.ix_(loc_1, loc_2)] += Xj

        # Preparing for loop reentry or exit
        Xj_topleft_1 += _floor(4 * M1) - _floor(2 * M1)
        Xj_topleft_2 += _floor(4 * M2) - _floor(2 * M2)

        lowpass = lowpass_next

    if is_real:
        Y = X
        X = np.rot90(X, 2)
        X = X + np.conj(Y)

    # Coarsest wavelet level
    M1 = M1 / 2
    M2 = M2 / 2
    Xj = fftshift(fft2(ifftshift(C[0][0]))) / np.sqrt(C[0][0].size)
    loc_1 = Xj_topleft_1 + np.arange(2 * _floor(4 * M1) + 1) - 1
    loc_2 = Xj_topleft_2 + np.arange(2 * _floor(4 * M2) + 1) - 1
    X[np.ix_(loc_1, loc_2)] += Xj * lowpass

    # Finest level
    M1 = N1 / 3
    M2 = N2 / 3
    if finest == 1:
        # Folding back onto N1-by-N2 matrix
        shift_1 = _floor(2 * M1) - _floor(N1 / 2)
        shift_2 = _floor(2 * M2) - _floor(N2 / 2)
        Y = X[:, np.arange(N2) + shift_2]
        Y[:, np.arange(N2 - shift_2, N2)] += X[:, :shift_2]
        Y[:, :shift_2] += X[:, N2 + shift_2:N2 + 2 * shift_2]
        X = Y[np.arange(N1) + shift_1, :]
        X[np.arange(N1 - shift_1, N1), :] += Y[:shift_1, :]
        X[:shift_1, :] += Y[N1 + shift_1:N1 + 2 * shift_1, :]
    else:
        # Extension to a N1-by-N2 matrix
        Y = fftshift(fft2(ifftshift(C[nbscales - 1][0]))) / \
            np.sqrt(C[nbscales - 1][0].size)
        X_topleft_1 = _ceil((N1 + 1) / 2) - _floor(M1)
        X_topleft_2 = _ceil((N2 + 1) / 2) - _floor(M2)
        loc_1 = X_topleft_1 + np.arange(2 * _floor(M1) + 1) - 1
        loc_2 = X_topleft_2 + np.arange(2 * _floor(M2) + 1) - 1
        Y[np.ix_(loc_1, loc_2)] = Y[np.ix_(loc_1, loc_2)] * hipass_finest + X
        X = Y

    x = fftshift(ifft2(ifftshift(X))) * np.sqrt(X.size)
    if is_real:
        x = np.real(x)

    return x



# Purpose: Module for handling Collection objects.
# Author: Minzhe Hu
# Date: 2025.3.30
# Email: hmz2018@mail.ustc.edu.cn
import os
import warnings
import numpy as np
from copy import deepcopy
from tqdm import tqdm
from glob import glob
from daspy.core.read import read
from daspy.core.dasdatetime import DASDateTime


cascade_method = ['time_integration', 'time_differential', 'downsampling',
                  'bandpass', 'bandstop', 'lowpass', 'highpass',
                  'lowpass_cheby_2']

class Collection(object):
    def __init__(self, fpath, ftype=None, flength=None, meta_from_file=True,
                 timeinfo_slice=slice(None), timeinfo_format=None, 
                 timeinfo_tz=None, timeinfo_from_basename=True, **kwargs):
        """
        :param fpath: str or Sequence of str. File path(s) containing data.
        :param ftype: None or str. None for automatic detection, or 'pkl',
            'pickle', 'tdms', 'h5', 'hdf5', 'segy', 'sgy', 'npy'.
        :param flength: float. The duration of a single file in senconds.
        :param meta_from_file: bool or 'all'. False for manually set dt, dx, fs
            and gauge_length. True for extracting dt, dx, fs and gauge_length
            from first 2 file. 'all' for exracting and checking these metadata
            from all file.
        :param timeinfo_slice: slice. Slice for extracting start time from file
            name.
        :param timeinfo_format: str. Format for extracting start time from file
            name.
        :param timeinfo_tz: datetime.timezone. Time zone for extracting start
            time from file name.
        :param timeinfo_from_basename: bool. If True, timeinfo_format will use
            DASDateTime.strptime to basename of fpath.
        :param nch: int. Channel number.
        :param nt: int. Sampling points of each file.
        :param dx: number. Channel interval in m.
        :param fs: number. Sampling rate in Hz.
        :param gauge_length: number. Gauge length in m.
        """
        if isinstance(fpath, (list, tuple)):
            self.flist = []
            for fp in fpath:
                self.flist.extend(glob(fp))
        else:
            self.flist = glob(fpath)
        if not len(self.flist):
            raise ValueError('No file input.')
        self.flist.sort()
        self.ftype = ftype
        for key in ['nch', 'nt', 'dx', 'fs', 'gauge_length']:
            if key in kwargs.keys():
                setattr(self, key, kwargs[key])
        if timeinfo_format is None and not meta_from_file:
            meta_from_file = True

        if meta_from_file == 'all':
            ftime = []
            metadata_list = []
            for f in self.flist:
                sec = read(f, ftype=ftype, headonly=True)
                if not hasattr(sec, 'gauge_length'):
                    sec.gauge_length = None
                ftime.append(sec.start_time)
                metadata_list.append((sec.nch, sec.nt, sec.dx, sec.fs,
                                      sec.gauge_length, sec.duration))

            if len(set(metadata_list)) > 1:
                warnings.warn('More than one kind of setting detected.')
            metadata = max(metadata_list, key=metadata_list.count)
            for i, key in enumerate(['nch', 'nt', 'dx', 'fs', 'gauge_length']):
                if not hasattr(self, key):
                    setattr(self, key, metadata[i])
                if flength is None:
                    flength = metadata[-1]
            self.ftime = ftime
        elif meta_from_file:
            i = int(len(self.flist) > 1)
            sec = read(self.flist[i], ftype=ftype, headonly=True)
            if timeinfo_format is None:
                if flength is None:
                    flength = sec.duration
                self.ftime = [sec.start_time + (j - i) * flength for j in
                            range(len(self))]
            if not hasattr(sec, 'gauge_length'):
                sec.gauge_length = None
            metadata = (sec.nch, sec.nt, sec.dx, sec.fs, sec.gauge_length)
            for i, key in enumerate(['nch', 'nt', 'dx', 'fs', 'gauge_length']):
                if not hasattr(self, key):
                    setattr(self, key, metadata[i])

        if not hasattr(self, 'ftime'):
            if timeinfo_from_basename:
                flist_use = [os.path.basename(f) for f in self.flist]
            else:
                flist_use = self.flist
            if timeinfo_tz is None:
                self.ftime = [DASDateTime.strptime(f[timeinfo_slice],
                    timeinfo_format) for f in flist_use]
            else:
                if '%z' in timeinfo_format.lower():
                    self.ftime = [DASDateTime.strptime(f[timeinfo_slice],
                        timeinfo_format).astimezone(timeinfo_tz) for f in
                        flist_use]
                else:
                    self.ftime = [DASDateTime.strptime(f[timeinfo_slice],
                        timeinfo_format).replace(tzinfo=timeinfo_tz) for f in
                        flist_use]

        self._sort()
        if flength is None:
            if len(self.flist) > 2:
                time_diff = np.round(np.diff(self.ftime[1:]).astype(float))
                flength_set, counts = np.unique(time_diff, return_counts=True)
                if len(flength_set) > 1:
                    warnings.warn('File start times are unevenly spaced. Data '
                                  'may not be continuous and self.flength may '
                                  'be incorrectly detected.')
                flength = flength_set[counts.argmax()]
            elif len(self.flist) == 2:
                flength = self.ftime[1] - self.ftime[0]
            else:
                flength = read(self.flist[0], ftype=ftype,
                               headonly=True).duration
        elif flength <= 0:
           raise ValueError('dt must > 0')
        
        self.flength = flength

    def __str__(self):
        if len(self) == 1:
            describe = f'       flist: {self.flist}\n'
        elif len(self) <= 5:
            describe = f'       flist: {len(self)} files\n' + \
                       f'              {self.flist}\n'
        else:
            describe = f'       flist: {len(self)} files\n' + \
                       f'              [{self[0]},\n' + \
                       f'               {self[1]},\n' + \
                       f'               ...,\n' + \
                       f'               {self[-1]}]\n'
            
        describe += f'       ftime: {self.start_time} to {self.end_time}\n' + \
                    f'     flength: {self.flength}\n'
        for key in ['nch', 'nt', 'dx', 'fs', 'gauge_length']:
            if hasattr(self, key):
                long_key = key.rjust(12)
                value = getattr(self, key)
                describe += f'{long_key}: {value}\n'

        return describe

    __repr__ = __str__

    def __getitem__(self, i):
        return self.flist[i]

    def __len__(self):
        return len(self.flist)

    def _sort(self):
        sort = np.argsort(self.ftime)
        self.ftime = [self.ftime[i] for i in sort]
        self.flist = [self.flist[i] for i in sort]
        return self

    @property
    def start_time(self):
        return self.ftime[0]

    @property
    def end_time(self):
        return self.ftime[-1] + self.flength

    @property
    def duration(self):
        return self.end_time - self.start_time

    @property
    def file_size(self):
        return os.path.getsize(self[1])

    def copy(self):
        return deepcopy(self)

    def select(self, stime=None, etime=None, readsec=False, **kwargs):
        """
        Select a period of data.

        :param stime, etime: DASDateTime. Start and end time of required data.
        :param readsec: bool. If True, read as a instance of daspy.Section and
            return. If False, update self.flist.
        :param ch1: int. The first channel required. Only works when
            readsec=True.
        :param ch2: int. The last channel required (not included). Only works
            when readsec=True.
        :param dch: int. Channel step. Only works when readsec=True.
        """
        if stime is None:
            stime = self.ftime[0]
        elif stime - self.ftime[0] < 0:
            warnings.warn('stime is earlier than the start time of the first '
                          'file. Set stime to self.ftime[0].')

        if etime is None:
            etime = self.ftime[-1] + self.flength
        elif etime - self.ftime[-1] > self.flength:
            warnings.warn('etime is later than the end time of the last file. '
                          'Set etime to self.ftime[-1] + self.flength.')

        if stime > etime:
            raise ValueError('Start time can\'t be later than end time.')

        flist = []
        ftime = []
        for i in range(len(self)):
            if (stime - self.flength) < self.ftime[i] < etime:
                flist.append(self.flist[i])
                ftime.append(self.ftime[i])

        if len(flist) == 0:
            warnings.warn('Out of collection time range.')
            return None
        if readsec:
            sec = read(flist[0], **kwargs)
            for f in flist[1:]:
                sec += read(f, **kwargs)
            sec.trimming(tmin=stime, tmax=etime)
            return sec
        else:
            self.flist = flist
            self.ftime = ftime
            return self

    def _optimize_for_continuity(self, operations):
        method_list = []
        kwargs_list = []
        if not isinstance(operations[0], (list, tuple)):
            operations = [operations]
        for opera in operations:
            method, kwargs = opera
            if method == 'downsampling':
                if (hasattr(kwargs, 'lowpass_filter') and not\
                    kwargs['lowpass_filter']) or not hasattr(kwargs, 'tint'):
                    method_list.append('downsampling')
                    kwargs_list.append(kwargs)
                else:
                    method_list.extend(['lowpass_cheby_2', 'downsampling'])
                    kwargs['lowpass_filter'] = False
                    kwargs0 = dict(freq=self.fs/2/kwargs['tint'], zi=0)
                    kwargs_list.extend([kwargs0, kwargs])
            else:
                if method in ['taper', 'cosine_taper']:
                    kwargs.setdefault('side', 'both')
                elif method in ['bandpass', 'bandstop', 'lowpass', 'highpass',
                                'lowpass_cheby_2']:
                    kwargs.setdefault('zi', 0)

                method_list.append(method)
                kwargs_list.append(kwargs)
        return method_list, kwargs_list

    def process(self, operations, savepath='./processed', merge=1,
                suffix='_pro', ftype=None, **read_kwargs):
        """
        :param operations: list. Each element of operations list should be [str
            of method name, dict of kwargs].
        :param savepath: str. Path to save processed files.
        :param merge: int or str. int for merge several processed files into 1.
            'all' for merge all files.
        :param suffix: str. Suffix for processed files.
        :param ftype: None or str. None for automatic detection, or 'pkl',
            'pickle', 'tdms', 'h5', 'hdf5', 'segy', 'sgy', 'npy'.
        :param read_kwargs: dict. Paramters for read function.
        """
        if not os.path.exists(savepath):
            os.makedirs(savepath)
        method_list, kwargs_list = self._optimize_for_continuity(operations)
        if merge == 'all' or merge > len(self):
            merge = len(self)
        for i in tqdm(range(0, len(self))):
            f = self[i]
            if os.path.getsize(f) == 0:
                for j, method in enumerate(method_list):
                    if method == 'time_integration':
                        kwargs_list[j]['c'] = 0
                    elif method == 'time_differential':
                        kwargs_list[j]['prepend'] = 0
                    elif method in ['bandpass', 'bandstop', 'lowpass',
                                    'highpass', 'lowpass_cheby_2']:
                        kwargs_list[j]['zi'] = 0
                continue
            sec = read(f, ftype=self.ftype, **read_kwargs)
            for j, method in enumerate(method_list):
                if method in ['taper', 'cosine_taper']:
                    if not ((i==0 and kwargs_list[j] != 'right') or
                             (i == len(self) - 1 and kwargs_list[j] != 'left')):
                        continue
                out = eval(f'sec.{method}')(**kwargs_list[j])
                if method == 'time_integration':
                    kwargs_list[j]['c'] = sec.data[:, -1]
                elif method == 'time_differential':
                    kwargs_list[j]['prepend'] = sec.data[:, -1]
                elif method in ['bandpass', 'bandstop', 'lowpass', 'highpass',
                                'lowpass_cheby_2']:
                    kwargs_list[j]['zi'] = out
            
            if i % merge == 0: 
                if i != 0:
                    sec_merge.save(filepath)
                sec_merge = sec
                f0, f1 = os.path.splitext(os.path.basename(f))
                if ftype is not None:
                    f1 = ftype
                filepath = os.path.join(savepath, f0+suffix+f1)
            else:
                sec_merge += sec
        sec_merge.save(filepath)



# Purpose: Module for handling DASDateTime objects.
# Author: Minzhe Hu
# Date: 2025.3.29
# Email: hmz2018@mail.ustc.edu.cn
import time
from typing import Iterable
from datetime import datetime, timedelta, timezone


utc = timezone.utc
local_tz = timezone(timedelta(seconds=-time.altzone))


class DASDateTime(datetime):
    def __add__(self, other):
        if isinstance(other, Iterable):
            out = []
            for t in other:
                out.append(self + t)
            return out
        elif not isinstance(other, timedelta):
            other = timedelta(seconds=float(other))
        return super().__add__(other)

    def __sub__(self, other):
        if isinstance(other, Iterable):
            out = []
            for t in other:
                out.append(self - t)
            return out
        elif isinstance(other, datetime):
            return datetime.__sub__(*self._unify_tz(other)).total_seconds()
        elif not isinstance(other, timedelta):
            other = timedelta(seconds=other)
        return super().__sub__(other)

    def __le__(self, other):
        return datetime.__le__(*self._unify_tz(other))

    def __lt__(self, other):
        return datetime.__lt__(*self._unify_tz(other))

    def __ge__(self, other):
        return datetime.__ge__(*self._unify_tz(other))

    def __gt__(self, other):
        return datetime.__gt__(*self._unify_tz(other))

    def _unify_tz(self, other: datetime):
        if self.tzinfo and (not other.tzinfo):
            return self, other.replace(tzinfo=self.tzinfo)
        elif (not self.tzinfo) and other.tzinfo:
            return self.replace(tzinfo=other.tzinfo), other
        return self, other

    def local(self):
        return self.astimezone(tz=local_tz)

    def utc(self):
        return self.astimezone(tz=utc)
    
    def remove_tz(self):
        return self.replace(tzinfo=None)

    @classmethod
    def from_datetime(cls, dt: datetime):
        return cls.fromtimestamp(dt.timestamp(), tz=dt.tzinfo)

    @classmethod
    def from_obspy_UTCDateTime(cls, dt):
        return cls.from_datetime(dt.datetime)

    def to_datetime(self):
        return datetime.fromtimestamp(self.timestamp(), tz=self.tzinfo)

    def to_obspy_UTCDateTime(self):
        from obspy import UTCDateTime
        return UTCDateTime(UTCDateTime(self.to_datetime()))

    @classmethod
    def strptime(cls, date_string, format):
        """
        string, format -> new datetime parsed from a string
        (like time.strptime()).
        """
        from _strptime import _strptime
        tt, fraction, gmtoff_fraction = _strptime(date_string, format)
        tzname, gmtoff = tt[-2:]
        args = tt[:6] + (fraction,)
        if gmtoff is not None:
            tzdelta = timedelta(seconds=gmtoff, microseconds=gmtoff_fraction)
            if tzname:
                tz = timezone(tzdelta, tzname)
            else:
                tz = timezone(tzdelta)
            args += (tz,)
        elif tt[-3] == 0:
            args += (utc,)

        return cls(*args)


# Purpose: Waveform decomposition
# Author: Minzhe Hu
# Date: 2024.5.13
# Email: hmz2018@mail.ustc.edu.cn
import numpy as np
from numpy.fft import irfft2, ifftshift
from daspy.basic_tools.preprocessing import padding, cosine_taper
from daspy.basic_tools.freqattributes import next_pow_2, fk_transform
from daspy.advanced_tools.denoising import curvelet_denoising


def fk_fan_mask(f, k, fmin=None, fmax=None, kmin=None, kmax=None, vmin=None,
                vmax=None, edge=0.1, flag=None):
    """
    Make a fan mask in f-k domain for f-k filter.

    :param f: Frequency sequence.
    :param k: Wavenumber sequence.
    :param fmin, fmax, kmin, kmax, vmin, vmax: float or or sequence of 2 floats.
        Sequence of 2 floats represents the start and end of taper.
    :param edge: float. The width of fan mask taper edge.
    :param flag: -1 keep only negative apparent velocities, 0 keep both postive
        and negative apparent velocities, 1 keep only positive apparent
        velocities.
    :return: Fan mask.
    """
    ff = np.tile(f, (len(k), 1))
    kk = np.tile(k, (len(f), 1)).T
    vv = - np.divide(ff, kk, out=np.ones_like(ff) * 1e10, where=kk != 0)
    mask = np.ones(vv.shape)
    for phy_quan in ['f', 'k', 'v']:
        p = eval(phy_quan * 2)
        pmin = eval(phy_quan + 'min')
        if pmin:
            if isinstance(pmin, (tuple, list, np.ndarray)):
                tp_b, tp_e = min(pmin), max(pmin)
            else:
                tp_b, tp_e = pmin * max(1 - edge / 2, 0), pmin * (1 + edge / 2)
            tp_wid = tp_e - tp_b
            mask[(abs(p) <= tp_b)] = 0
            area = (abs(p) > tp_b) & (abs(p) < tp_e)
            mask[area] *= 0.5 - 0.5 * \
                np.cos(((abs(p[area]) - tp_b) / tp_wid) * np.pi)

        pmax = eval(phy_quan + 'max')
        if pmax:
            if isinstance(pmax, (tuple, list, np.ndarray)):
                tp_b, tp_e = max(pmax), min(pmax)
            else:
                tp_b, tp_e = pmax * (1 + edge / 2), pmax * (1 - edge / 2)
            tp_wid = tp_b - tp_e
            mask[(abs(p) >= tp_b)] = 0
            area = (abs(p) > tp_e) & (abs(p) < tp_b)
            mask[area] *= 0.5 - 0.5 * \
                np.cos(((tp_b - abs(p[area])) / tp_wid) * np.pi)

    if flag:
        mask[np.sign(vv) == flag] = 0
    return mask


def fk_filter(data, dx, fs, taper=(0.02, 0.05), pad='default', mode='decompose',
              fmin=None, fmax=None, kmin=None, kmax=None, vmin=None, vmax=None,
              edge=0.1, flag=None, verbose=False):
    """
    Transform the data to the f-k domain using 2-D Fourier transform method, and
    transform back to the x-t domain after filtering.

    :param data: numpy.ndarray. Data to do fk filter.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param taper: float or sequence of floats. Each float means decimal
        percentage of Tukey taper for corresponding dimension (ranging from 0 to
        1). Default is 0.1 which tapers 5% from the beginning and 5% from the
        end.
    :param pad: Pad the data or not. It can be float or sequence of floats. Each
        float means padding percentage before FFT for corresponding dimension.
        If set to 0.1 will pad 5% before the beginning and after the end.
        'default' means pad both dimensions to next power of 2. None or False
        means don't pad data before or during Fast Fourier Transform.
    :param mode: str. 'remove' for denoising, 'retain' for extraction, and
        'decompose' for decomposition.
    :param fmin, fmax, kmin, kmax, vmin, vmax: float or or sequence of 2 floats.
        Sequence of 2 floats represents the start and end of taper.
    :param edge: float. The width of fan mask taper edge.
    :param flag: -1 keep only negative apparent velocities, 0 keep both postive
        and negative apparent velocities, 1 keep only positive apparent
        velocities.
    :param verbose: If True, return filtered data, f-k spectrum, frequency
        sequence, wavenumber sequence and f-k mask.
    :return: Filtered data and some variables in the process if verbose==True.
    """
    data_tp = cosine_taper(data, taper)
    if pad == 'default':
        nch, nt = data.shape
        dn = (next_pow_2(nch) - nch, next_pow_2(nt) - nt)
        nfft = None
    elif pad is None or pad is False:
        dn = 0
        nfft = None
    else:
        dn = np.round(np.array(pad) * data.shape).astype(int)
        nfft = 'default'

    data_pd = padding(data_tp, dn)
    nch, nt = data_pd.shape

    fk, f, k = fk_transform(data_pd, dx, fs, taper=0, nfft=nfft)

    mask = fk_fan_mask(f, k, fmin, fmax, kmin, kmax, vmin, vmax, edge=edge,
                       flag=flag)

    if mode == 'remove':
        mask = 1 - mask

    if mode == 'decompose':
        data_flt1 = irfft2(ifftshift(fk * mask, axes=0)).real[:nch, :nt]
        data_flt1 = padding(data_flt1, dn, reverse=True)
        data_flt2 = irfft2(ifftshift(fk * (1 - mask), axes=0)).real[:nch, :nt]
        data_flt2 = padding(data_flt2, dn, reverse=True)
        if verbose:
            return data_flt1, data_flt2, fk, f, k, mask
        else:
            return data_flt1, data_flt2
    else:
        data_flt = irfft2(ifftshift(fk * mask, axes=0)).real[:nch, :nt]
        data_flt = padding(data_flt, dn, reverse=True)
        if verbose:
            return data_flt, fk, f, k, mask
        else:
            return data_flt


def curvelet_windowing(data, dx, fs, mode='decompose', vmin=0, vmax=np.inf,
                       flag=None, pad=0.3, scale_begin=3, nbscales=None,
                       nbangles=16, finest=1):
    """
    Use curevelet transform to keep cooherent signal with certain velocity
    range. {Atterholt et al., 2022 , Geophys. J. Int.}

    :param data: numpy.ndarray. Data to decomposite.
    :param dx: Channel interval in m.
    :param fs: Sampling rate in Hz.
    :param mode: str. 'remove' for denoising, 'retain' for extraction, and
        'decompose' for decomposition.
    :param vmin, vmax: float. Velocity range in m/s.
    :param flag: -1 keep only negative apparent velocities, 0 keep both postive
        and negative apparent velocities, 1 keep only positive apparent
        velocities.
    :param pad: float or sequence of floats. Each float means padding percentage
        before FFT for corresponding dimension. If set to 0.1 will pad 5% before
        the beginning and after the end.
    :param scale_begin: int. The beginning scale to do coherent denoising.
    :param nbscales: int. Number of scales including the coarsest wavelet level.
        Default set to ceil(log2(min(M,N)) - 3).
    :param nbangles: int. Number of angles at the 2nd coarsest level,
        minimum 8, must be a multiple of 4.
    :param finest: int. Objects at the finest scale. 1 for curvelets, 2 for
        wavelets. Curvelets are more precise while wavelets are more efficient.
    :return: numpy.ndarray. Decomposed data.
    """
    return curvelet_denoising(data, choice=1, pad=pad, vmin=vmin, vmax=vmax,
                              flag=flag, dx=dx, fs=fs, mode=mode,
                              scale_begin=scale_begin, nbscales=nbscales,
                              nbangles=nbangles, finest=finest)
